from smrf.distribute import image_data
from smrf.envphys.constants import IR_WAVELENGTHS, VISIBLE_WAVELENGTHS
from smrf.envphys.solar import cloud, toporad, vegetation
from smrf.utils import utils


class Solar(image_data.image_data):
    """
    The :mod:`~smrf.distribute.solar.solar` class allows for variable specific
    distributions that go beyond the base class.

    Multiple steps are required to estimate solar radiation:

    1. Terrain corrected clear sky radiation
    2. Adjust solar radiation for vegetation effects
    3. Calculate net radiation using the albedo

    The Image Processing Workbench (IPW) includes a utility ``stoporad`` to
    model terrain corrected clear sky radiation over the DEM. Within
    ``stoporad``, the radiation transfer model ``twostream`` simulates the
    clear sky radiation on a flat surface for a range of wavelengths through
    the atmosphere :cite:`Dozier:1980` :cite:`Dozier&Frew:1981`
    :cite:`Dubayah:1994`. Terrain correction using the DEM adjusts for terrain
    shading and splits the clear sky radiation into beam and diffuse radiation.

    The second step requires sites measuring solar radiation. The measured
    solar radiation is compared to the modeled clear sky radiation from
    ``twostream``. The cloud factor is then the measured incoming solar
    radiation divided by the modeled radiation.  The cloud factor can be
    computed on an hourly timescale if the measurement locations are of high
    quality. For stations that are less reliable, we recommend calculating a
    daily cloud factor which divides the daily integrated measured radiation by
    the daily integrated modeled radiation.  This helps to reduce the problems
    that may be encountered from instrument shading, instrument calibration, or
    a time shift in the data. The calculated cloud factor at each station can
    then be distrubted using any of the method available in
    :mod:`smrf.spatial`. Since the cloud factor is not explicitly controlled
    by elevation like other variables, the values may be distributed without
    detrending to elevation. The modeled clear sky radiation (both beam and
    diffuse)  are adjusted for clouds using
    :mod:`smrf.envphys.radiation.cf_cloud`.

    The third step adjusts the cloud corrected solar radiation for vegetation
    affects, following the methods developed by Link and Marks (1999)
    :cite:`Link&Marks:1999`. The direct beam radiation  is corrected by:

    .. math::
        R_b = S_b * exp( -\\mu h / cos \\theta )

    where :math:`S_b` is the above canopy direct radiation, :math:`\\mu` is
    the extinction coefficient (:math:`m^{-1}`), :math:`h` is the canopy height
    (:math:`m`), :math:`\\theta` is the solar zenith angle, and :math:`R_b` is
    the canopy adjusted direct radiation. Adjusting the diffuse radiation is
    performed by:

    .. math::
        R_d = \\tau * R_d

    where :math:`R_d` is the diffuse adjusted radiation, :math:`\\tau` is the
    optical transmissivity of the canopy, and :math:`R_d` is the above canopy
    diffuse radiation. Values for :math:`\\mu` and :math:`\\tau` can be found
    in Link and Marks (1999) :cite:`Link&Marks:1999`, measured at study sites
    in Saskatchewan and Manitoba.

    The final step for calculating the net solar radiation requires the surface
    albedo from :mod:`smrf.distribute.albedo`. The net radiation is the sum of
    the of beam and diffuse canopy adjusted radiation multipled by one minus
    the albedo.

    Args:
        config: full configuration dictionary contain at least the sections
                albedo, and solar
        topo: Topo class :mod:`smrf.data.loadTopo.Topo`

    Attributes:
        albedoConfig: configuration from [albedo] section
        config: configuration from [albedo] section
        clear_ir_beam: numpy array modeled clear sky infrared beam radiation
        clear_ir_diffuse: numpy array modeled clear sky infrared diffuse
            radiation
        clear_vis_beam: numpy array modeled clear sky visible beam radiation
        clear_vis_diffuse: numpy array modeled clear sky visible diffuse
            radiation
        cloud_factor: numpy array distributed cloud factor
        cloud_ir_beam: numpy array cloud adjusted infrared beam radiation
        cloud_ir_diffuse: numpy array cloud adjusted infrared diffuse radiation
        cloud_vis_beam: numpy array cloud adjusted visible beam radiation
        cloud_vis_diffuse: numpy array cloud adjusted visible diffuse radiation
        ir_file: temporary file from ``stoporad`` for infrared clear sky
            radiation
        metadata: metadata for the station data
        net_solar: numpy array for the calculated net solar radiation
        stations: stations to be used in alphabetical order
        veg_height: numpy array of vegetation heights from
            :mod:`smrf.data.loadTopo.Topo`
        veg_ir_beam: numpy array vegetation adjusted infrared beam radiation
        veg_ir_diffuse: numpy array vegetation adjusted infrared diffuse
            radiation
        veg_k: numpy array of vegetation extinction coefficient from
            :mod:`smrf.data.loadTopo.Topo`
        veg_tau: numpy array of vegetation optical transmissivity from
            :mod:`smrf.data.loadTopo.Topo`
        veg_vis_beam: numpy array vegetation adjusted visible beam radiation
        veg_vis_diffuse: numpy array vegetation adjusted visible diffuse
            radiation
        vis_file: temporary file from ``stoporad`` for visible clear sky
            radiation

    """

    variable = 'solar'

    # these are variables that can be output
    OUTPUT_VARIABLES = {
        'clear_ir_beam': {
            'units': 'watt/m2',
            'standard_name': 'clear_sky_infrared_beam',
            'long_name': 'Clear sky infrared beam solar radiation'
        },
        'clear_ir_diffuse': {
            'units': 'watt/m2',
            'standard_name': 'clear_sky_infrared_diffuse',
            'long_name': 'Clear sky infrared diffuse solar radiation'
        },
        'clear_vis_beam': {
            'units': 'watt/m2',
            'standard_name': 'clear_sky_visible_beam',
            'long_name': 'Clear sky visible beam solar radiation'
        },
        'clear_vis_diffuse': {
            'units': 'watt/m2',
            'standard_name': 'clear_sky_visible_diffuse',
            'long_name': 'Clear sky visible diffuse solar radiation'
        },
        'cloud_ir_beam': {
            'units': 'watt/m2',
            'standard_name': 'cloud_infrared_beam',
            'long_name': 'Cloud corrected infrared beam solar radiation'
        },
        'cloud_ir_diffuse': {
            'units': 'watt/m2',
            'standard_name': 'cloud_infrared_diffuse',
            'long_name': 'Cloud corrected infrared diffuse solar radiation'
        },
        'cloud_vis_beam': {
            'units': 'watt/m2',
            'standard_name': 'cloud_visible_beam',
            'long_name': 'Cloud corrected visible beam solar radiation'
        },
        'cloud_vis_diffuse': {
            'units': 'watt/m2',
            'standard_name': 'cloud_visible_diffuse',
            'long_name': 'Cloud corrected visible diffuse solar radiation'
        },
        'net_solar': {
            'units': 'watt/m2',
            'standard_name': 'net_solar_radiation',
            'long_name': 'Net solar radiation'
        },
        'veg_ir_beam': {
            'units': 'watt/m2',
            'standard_name': 'vegetation_infrared_beam',
            'long_name': 'Vegetation corrected infrared beam solar radiation'
        },
        'veg_ir_diffuse': {
            'units': 'watt/m2',
            'standard_name': 'vegetation_infrared_diffuse',
            'long_name': 'Vegetation corrected infrared diffuse solar \
                radiation'
        },
        'veg_vis_beam': {
            'units': 'watt/m2',
            'standard_name': 'vegetation_visible_beam',
            'long_name': 'Vegetation corrected visible beam solar radiation'
        },
        'veg_vis_diffuse': {
            'units': 'watt/m2',
            'standard_name': 'vegetation_visible_diffuse',
            'long_name': 'Vegetation corrected visible diffuse solar radiation'
        }
    }

    CLEAR_SKY_THREAD_VARIABLES = [
        'clear_ir_beam',
        'clear_ir_diffuse',
        'clear_vis_beam',
        'clear_vis_diffuse'
    ]

    VEG_THREAD_VARIABLES = [
        'veg_ir_beam',
        'veg_ir_diffuse',
        'veg_vis_beam',
        'veg_vis_diffuse'
    ]

    CLOUD_THREAD_VARIABLES = [
        'cloud_ir_beam',
        'cloud_ir_diffuse',
        'cloud_vis_beam',
        'cloud_vis_diffuse'
    ]

    BASE_THREAD_VARIABLES = frozenset(
        CLEAR_SKY_THREAD_VARIABLES +
        CLOUD_THREAD_VARIABLES +
        VEG_THREAD_VARIABLES +
        ['net_solar']
    )

    # These are variables that are operate at the end only and do not need to
    # be written during main distribute loop
    post_process_variables = {}

    def __init__(self, config, topo):

        # extend the base class
        image_data.image_data.__init__(self, self.variable)

        self.config = config["solar"]
        self.albedoConfig = config["albedo"]

        self.topo = topo

        self._logger.debug('Created distribute.solar')

    def initialize(self, topo, data, date_time=None):
        """
        Initialize the distribution, soley calls
        :mod:`smrf.distribute.image_data.image_data._initialize`. Sets the
        following attributes:

        * :py:attr:`veg_height`
        * :py:attr:`veg_tau`
        * :py:attr:`veg_k`

        Args:
            topo: :mod:`smrf.data.loadTopo.Topo` instance contain topographic
                data and infomation
            data: data Pandas dataframe containing the station data,
                from :mod:`smrf.data.loadData` or :mod:`smrf.data.loadGrid`

        """

        self._logger.debug('Initializing distribute.solar')
        self.date_time = date_time
        # Solar has no stations. Relies on Cloud factor
        self.stations = None
        self._initialize(topo, data.metadata)
        self.veg_height = topo.veg_height
        self.veg_tau = topo.veg_tau
        self.veg_k = topo.veg_k

    def distribute(self, date_time, cloud_factor, illum_ang, cosz, azimuth,
                   albedo_vis, albedo_ir):
        """
        Distribute air temperature given a Panda's dataframe for a single time
        step. Calls :mod:`smrf.distribute.image_data.image_data._distribute`.

        If the sun is up, i.e. ``cosz > 0``, then the following steps are
        performed:

        1. Model clear sky radiation
        2. Cloud correct with :mod:`!smrf.distribute.solar.solar.cloud_correct`
        3. vegetation correct with
            :mod:`!smrf.distribute.solar.solar.veg_correct`
        4. Calculate net radiation with
            :mod:`!smrf.distribute.solar.solar.calc_net`

        If sun is down, then all calculated values will be set to ``None``,
        signaling the output functions to put zeros in their place.

        Args:
            cloud_factor: Numpy array of the domain for cloud factor
            cosz: cosine of the zenith angle for the basin, from
                :mod:`smrf.envphys.radiation.sunang`
            azimuth: azimuth to the sun for the basin, from
                :mod:`smrf.envphys.radiation.sunang`
            albedo_vis: numpy array for visible albedo, from
                :mod:`smrf.distribute.albedo.Albedo.albedo_vis`
            albedo_ir: numpy array for infrared albedo, from
                :mod:`smrf.distribute.albedo.Albedo.albedo_ir`

        """

        self._logger.debug(f'{date_time} Distributing solar')

        # Only calculate solar if the sun is up
        if cosz > 0:
            self.cloud_factor = cloud_factor.copy()

            # --------------------------------------------
            # calculate clear sky radiation

            # Not all the clean but it will work for now
            val_beam, val_diffuse = self.calc_stoporad(
                date_time, illum_ang, cosz, azimuth,
                albedo_ir, IR_WAVELENGTHS)

            setattr(self, 'clear_ir_beam', val_beam)
            setattr(self, 'clear_ir_diffuse', val_diffuse)
            self.ir_beam = val_beam.copy()
            self.ir_diffuse = val_diffuse.copy()

            val_beam, val_diffuse = self.calc_stoporad(
                date_time, illum_ang, cosz, azimuth,
                albedo_vis, VISIBLE_WAVELENGTHS)

            setattr(self, 'clear_vis_beam', val_beam)
            setattr(self, 'clear_vis_diffuse', val_diffuse)
            self.vis_beam = val_beam.copy()
            self.vis_diffuse = val_diffuse.copy()

            # --------------------------------------------
            # correct clear sky for cloud
            if self.config['correct_cloud']:
                self.cloud_correct()
                # copy output for output variables
                self.cloud_vis_beam = self.vis_beam.copy()
                self.cloud_vis_diffuse = self.vis_diffuse.copy()
                self.cloud_ir_beam = self.ir_beam.copy()
                self.cloud_ir_diffuse = self.ir_diffuse.copy()

            # --------------------------------------------
            # correct cloud for veg
            if self.config['correct_veg']:
                self.veg_correct(illum_ang)
                self.veg_vis_beam = self.vis_beam.copy()
                self.veg_vis_diffuse = self.vis_diffuse.copy()
                self.veg_ir_beam = self.ir_beam.copy()
                self.veg_ir_diffuse = self.ir_diffuse.copy()

            # --------------------------------------------
            # calculate net radiation
            self.calc_net(albedo_vis, albedo_ir)

        else:

            self._logger.debug('Sun is down, see you in the morning!')

            # clear sky
            self.clear_vis_beam = None
            self.clear_vis_diffuse = None
            self.clear_ir_beam = None
            self.clear_ir_diffuse = None

            # cloud
            self.cloud_vis_beam = None
            self.cloud_vis_diffuse = None
            self.cloud_ir_beam = None
            self.cloud_ir_diffuse = None

            # canopy
            self.veg_vis_beam = None
            self.veg_vis_diffuse = None
            self.veg_ir_beam = None
            self.veg_ir_diffuse = None

            # net
            self.net_solar = None

    def distribute_thread(self, smrf_queue, data_queue=None):
        """
        Distribute the data using threading. All data is provided and
        ``distribute_thread`` will go through each time step following the
        methods outlined in :mod:`smrf.distribute.solar.solar.distribute`. The
        data smrf_queues puts the distributed data into:

        * :py:attr:`net_solar`

        Args:
            smrf_queue: smrf_queue dictionary for all variables
            data: pandas dataframe for all data, indexed by date time
        """
        self._logger.info("Distributing {}".format(self.variable))
        for date_time in self.date_time:

            cosz = smrf_queue['cosz'].get(date_time)
            azimuth = smrf_queue['azimuth'].get(date_time)
            illum_ang = smrf_queue['illum_ang'].get(date_time)
            albedo_ir = smrf_queue['albedo_ir'].get(date_time)
            albedo_vis = smrf_queue['albedo_vis'].get(date_time)
            self.cloud_factor = smrf_queue['cloud_factor'].get(date_time)

            self.distribute(
                date_time,
                self.cloud_factor,
                illum_ang,
                cosz,
                azimuth,
                albedo_vis,
                albedo_ir)

            for cstv in self.CLEAR_SKY_THREAD_VARIABLES:
                smrf_queue[cstv].put([date_time, getattr(self, cstv)])

            # Add the cloud corrected variables to the smrf_queue
            if self.config['correct_cloud']:
                for vtv in self.VEG_THREAD_VARIABLES:
                    smrf_queue[vtv].put([date_time, getattr(self, vtv)])

            # Add the veg correct variables to the smrf_queue
            if self.config['correct_veg']:
                for ctv in self.CLOUD_THREAD_VARIABLES:
                    smrf_queue[ctv].put([date_time, getattr(self, ctv)])

            smrf_queue['net_solar'].put([date_time, self.net_solar])

    def cloud_correct(self):
        """
        Correct the modeled clear sky radiation for cloud cover using
        :mod:`smrf.envphys.radiation.cloud.cf_cloud`.
        Sets :py:attr:`cloud_vis_beam` and :py:attr:`cloud_vis_diffuse`.
        """

        self._logger.debug('Correcting clear sky radiation for clouds')
        self.vis_beam, self.vis_diffuse = cloud.cf_cloud(
            self.vis_beam,
            self.vis_diffuse,
            self.cloud_factor)

        self.ir_beam, self.ir_diffuse = cloud.cf_cloud(
            self.ir_beam,
            self.ir_diffuse,
            self.cloud_factor)

    def veg_correct(self, illum_ang):
        """
        Correct the cloud adjusted radiation for vegetation using
        :mod:`smrf.envphys.radiation.vegetation.veg_beam` and
        :mod:`smrf.envphys.radiation.vegetation.veg_diffuse`. Sets
        :py:attr:`veg_vis_beam`, :py:attr:`veg_vis_diffuse`,
        :py:attr:`veg_ir_beam`, and :py:attr:`veg_ir_diffuse`.

        Args:
            illum_ang: numpy array of the illumination angle over the DEM, from
                :mod:`smrf.envphys.radiation.sunang`

        """

        self._logger.debug('Correcting radiation for vegetation')

        # calculate for visible
        # correct beam
        self.vis_beam = vegetation.solar_veg_beam(
            self.vis_beam,
            self.veg_height,
            illum_ang,
            self.veg_k)

        # correct diffuse
        self.vis_diffuse = vegetation.solar_veg_diffuse(
            self.vis_diffuse,
            self.veg_tau)

        # calculate for ir
        # correct beam
        self.ir_beam = vegetation.solar_veg_beam(
            self.ir_beam,
            self.veg_height,
            illum_ang,
            self.veg_k)

        # correct diffuse
        self.ir_diffuse = vegetation.solar_veg_diffuse(
            self.ir_diffuse,
            self.veg_tau)

    def calc_net(self, albedo_vis, albedo_ir):
        """
        Calculate the net radiation using the vegetation adjusted radiation.
        Sets :py:attr:`net_solar`.

        Args:
            albedo_vis: numpy array for visible albedo, from
                :mod:`smrf.distribute.albedo.Albedo.albedo_vis`
            albedo_ir: numpy array for infrared albedo, from
                :mod:`smrf.distribute.albedo.Albedo.albedo_ir`
        """

        self._logger.debug('Calculating net radiation')

        # calculate net visible
        vv_n = (self.vis_beam + self.vis_diffuse) * (1 - albedo_vis)
        vv_n = utils.set_min_max(vv_n, self.min, self.max)

        # calculate net ir
        vir_n = (self.ir_beam + self.ir_diffuse) * (1 - albedo_ir)
        vir_n = utils.set_min_max(vir_n, self.min, self.max)

        # calculate total net
        self.net_solar = vv_n + vir_n
        self.net_solar = utils.set_min_max(self.net_solar, self.min, self.max)

    def calc_stoporad(self, date_time, illum_ang, cosz, azimuth,
                      albedo_surface, wavelength_range=VISIBLE_WAVELENGTHS):
        """Run stoporad for the given date_time and wavelength range

        Args:
            date_time (datetime): datetime object
            illum_ang (np.array): numpy array of cosing of local illumination
                angles
            cosz (float): cosine of the zenith angle for the basin
            azimuth (float): azimuth to the sun for the basin
            albedo_surface (np.array): albedo should match wavelengths
                specified
            wavelength_range (list, optional): wavelengths to integrate over.
                Defaults to [0.28, 0.7].

        Returns:
            tuple: clear sky beam and diffuse radiation
        """

        clear_beam, clear_diffuse = toporad.stoporad(
            date_time,
            self.topo,
            cosz,
            azimuth,
            illum_ang,
            albedo_surface,
            wavelength_range=wavelength_range,
            tau_elevation=self.config['clear_opt_depth'],
            tau=self.config['clear_tau'],
            omega=self.config['clear_omega'],
            scattering_factor=self.config['clear_gamma'])

        return clear_beam, clear_diffuse



"""
Create classes for running on multiple threads

20160323 Scott Havens
"""

import logging
import sys
import threading
from queue import Empty, Full, Queue
from time import time as _time

import numpy as np


class DateQueueThreading(Queue):
    """
    DateQueue extends queue.Queue module
    Stores the items in a dictionary with date_time keys
    When values are retrieved, it will not remove them and will
    require cleaning at the end to not have to many values

    20160323 Scott Havens
    """

    def __init__(self, maxsize=0, timeout=None, name=None):
        # extend the base class
        Queue.__init__(self, maxsize)
        self.timeout = timeout
        logger_name = __name__
        self.name = name
        if name is not None:
            logger_name = '{}.{}'.format(logger_name, name)

        self._logger = logging.getLogger(logger_name)

    # The following override the methods in Queue to use
    # a date approach to the queue

    # Initialize the queue representation
    def _init(self, maxsize):
        self.queue = {}

    def _qsize(self, len=len):
        return len(self.queue)

    # Put a new item in the queue
    def _put(self, item):
        # self.queue.append(item)
        self.queue.update({item[0]: item[1]})

    # Get an item from the queue
    def _get(self, index):
        return self.queue[index]

    def clean(self, index):
        """
        Need to clean it out so mimic the original get
        """
        self.not_empty.acquire()
        try:
            if index in self.queue:
                del self.queue[index]
                self.not_full.notifyAll()
        finally:
            self.not_empty.release()

    def get(self, index, block=True, timeout=None):
        """
        Remove and return an item from the queue.

        If optional args 'block' is true and 'timeout' is None (the default),
        block if necessary until an item is available. If 'timeout' is
        a non-negative number, it blocks at most 'timeout' seconds and raises
        the Empty exception if no item was available within that time.
        Otherwise ('block' is false), return an item if one is immediately
        available, else raise the Empty exception ('timeout' is ignored
        in that case).

        This is from queue.Queue but with modifcation for supplying what to get

        Args:
            index: datetime object representing the date/time being processed
            block: whether to wait for a variable to become available
            timeout: Seconds to wait before dropping error, none is forever.
        """
        timed_out = False
        # see if timeout was passed
        if timeout is not None:
            self.timeout = timeout

        self.not_empty.acquire()

        try:
            if not block:
                if not self._qsize():
                    raise Empty

            # No timeout specified.
            elif self.timeout is None:
                while index not in self.queue.keys():
                    self.not_empty.wait()

            #  Timeout in seconds was specified
            else:
                endtime = _time() + self.timeout

                while index not in self.queue.keys():
                    self.remaining = endtime - _time()
                    # Time out has occurred
                    if self.remaining <= 0.0:
                        self._logger.error(
                            "Timeout occurred while retrieving"
                            " an item at {} from queue".format(index))
                        timed_out = True
                        raise Empty

                    self.not_empty.wait(self.remaining)

            item = self._get(index)

            self.not_full.notifyAll()
            return item

        except Exception as e:
            self._logger.error(e)

        finally:
            self.not_empty.release()
            if timed_out:
                sys.exit()

    def put(self, item, block=True, timeout=None):
        """Put an item into the queue.

        If optional args 'block' is true and 'timeout' is None (the default),
        block if necessary until a free slot is available. If 'timeout' is
        a non-negative number, it blocks at most 'timeout' seconds and raises
        the Full exception if no free slot was available within that time.
        Otherwise ('block' is false), put an item on the queue if a free slot
        is immediately available, else raise the Full exception ('timeout'
        is ignored in that case).
        """
        timed_out = False

        # see if timeout was passed
        timeout = None
        if timeout is not None:
            self.timeout = timeout

        self.not_full.acquire()

        try:
            # Is there anything in the Queue
            if self.maxsize > 0:

                # Working with a thread set to block other threads
                # until complete
                if not block:
                    if self._qsize() == self.maxsize:
                        raise Full

                # Never timeout
                elif self.timeout is None:
                    while self._qsize() == self.maxsize:
                        self.not_full.wait()

                # Timeout is set
                else:
                    endtime = _time() + self.timeout

                    while self._qsize() == self.maxsize:
                        self.remaining = endtime - _time()

                        if self.remaining <= 0.0:
                            self._logger.error(
                                "Timeout occurred while putting"
                                " {} in the queue.".format(item[0]))
                            timed_out = True
                            raise Full

                        self.not_full.wait(self.remaining)
            self._put(item)

            self.unfinished_tasks += 1
            self.not_empty.notifyAll()

        except Exception as e:
            self._logger.error(e)

        finally:
            self.not_full.release()
            if timed_out:
                sys.exit()


class QueueCleaner(threading.Thread):
    """
    QueueCleaner that will go through all the queues and check
    if they all have a date in common.  When this occurs, all the
    threads will have processed that time step and it's not longer needed
    """

    def __init__(self, date_time, queue):
        """
        Args:
            date_time: array of date_time
            queue: dict of the queue
        """
        threading.Thread.__init__(self, name='cleaner')
        self.queues = queue
        self.date_time = date_time
        lname = "{}.{}".format(__name__, 'clean')
        self._logger = logging.getLogger(lname)
        self._logger.debug('Initialized cleaner thread')

    def run(self):
        """
        Go through the date times and look for when all the queues
        have that date_time
        """

        for t in self.date_time:

            # first do a get on all the data, this will ensure that
            # there is data there to be had
            for v in self.queues.keys():
                self.queues[v].get(t)

            # now that we know there is data in all of the queues
            # that have the same time, clean up those times
            for v in self.queues.keys():
                self.queues[v].clean(t)

            self._logger.debug('%s Cleaned from queues' % t)


class QueueOutput(threading.Thread):
    """
    Takes values from the queue and outputs using 'out_func'
    """

    def __init__(self, queue, date_time, out_func, out_frequency, nx, ny):
        """
        Args:
            date_time: array of date_time
            queue: dict of the queue
        """

        threading.Thread.__init__(self, name='output')
        self.queues = queue
        self.date_time = date_time
        self.out_func = out_func
        self.out_frequency = out_frequency
        self.nx = nx
        self.ny = ny
        lname = "{}.{}".format(__name__, 'output')
        self._logger = logging.getLogger(lname)
        self._logger.debug('Initialized output thread')

    def run(self):
        """
        Output the desired variables to a file.

        Go through the date times and look for when all the queues
        have that date_time
        """

        for output_count, t in enumerate(self.date_time):

            # output at the frequency and the last time step
            if (output_count % self.out_frequency == 0) or \
               (output_count == len(self.date_time)):

                # get the output variables then pass to the function
                for v in self.out_func.variable_dict.values():
                    if v['variable'] in self.queues.keys():
                        data = self.queues[v['variable']].get(t)

                        if data is None:
                            data = np.zeros((self.ny, self.nx))

                        # output the time step
                        self._logger.debug(
                            "{} threaded output for {}".format(
                                t, v['variable']))
                        self.out_func.output(v['variable'], data, t)

                    else:
                        self._logger.warning(
                            '{0} Output variable {1} not in queue'
                            .format(t, v['variable']))

                # put the value into the output queue so clean knows it's done
                self.queues['output'].put([t, True])

                self._logger.debug('%s Variables output from queues' % t)




import logging

import numpy as np

from smrf.distribute import image_data
from smrf.distribute.wind.wind_ninja import WindNinjaModel
from smrf.distribute.wind.winstral import WinstralWindModel
from smrf.utils import utils


class Wind(image_data.image_data):
    """
    The :mod:`~smrf.distribute.wind.wind` class allows for variable specific
    distributions that go beyond the base class.

    Three distribution methods are available for the Wind class:

    1. Winstral and Marks 2002 method for maximum upwind slope (maxus)
    2. Import WindNinja simulations
    3. Standard interpolation

    Args:
        self.config: The full SMRF configuration file

    Attributes:
        config: configuration from [wind] section
        wind_speed: numpy matrix of the wind speed
        wind_direction: numpy matrix of the wind direction
        veg_type: numpy array for the veg type, from
            :py:attr:`smrf.data.loadTopo.Topo.veg_type`
        _maxus_file: the location of the maxus NetCDF file
        maxus: the loaded library values from :py:attr:`_maxus_file`
        maxus_direction: the directions associated with the :py:attr:`maxus`
            values
        min: minimum value of wind is 0.447
        max: maximum value of wind is 35
        stations: stations to be used in alphabetical order

    """

    VARIABLE = 'wind'

    # these are variables that can be output
    OUTPUT_VARIABLES = {
        'flatwind': {
            'units': 'm/s',
            'standard_name': 'flatwind_wind_speed',
            'long_name': 'Simulated wind on a flat surface'
        },
        'wind_speed': {
            'units': 'm/s',
            'standard_name': 'wind_speed',
            'long_name': 'Wind speed'
        },
        'wind_direction': {
            'units': 'degrees',
            'standard_name': 'wind_direction',
            'long_name': 'Wind direction'
        }
    }
    # these are variables that are operate at the end only and do not need to
    # be written during main distribute loop
    post_process_variables = {}

    BASE_THREAD_VARIABLES = frozenset([
        'wind_speed',
        'wind_direction'
    ])

    def __init__(self, config):

        # extend the base class
        image_data.image_data.__init__(self, self.VARIABLE)
        self._logger = logging.getLogger(__name__)

        # check and assign the configuration
        self.smrf_config = config
        self.getConfig(config['wind'])

        if self.check_wind_model_type('interp'):
            # Straight interpolation of the wind
            self.wind_model = self
            self.wind_model.flatwind = None
            self.wind_model.cellmaxus = None
            self.wind_model.dir_round_cell = None

        elif self.check_wind_model_type('wind_ninja'):
            self.wind_model = WindNinjaModel(self.smrf_config)

        elif self.check_wind_model_type('winstral'):
            self.wind_model = WinstralWindModel(self.smrf_config)

        self._logger.debug('Created distribute.wind')

    def check_wind_model_type(self, wind_model):
        """Check if the wind model is of a given type

        Args:
            wind_model (str): name of the wind model

        Returns:
            bool: True/False if the wind_model matches the config
        """

        return self.config['wind_model'] == wind_model

    def initialize(self, topo, data, date_time=None):
        """
        Initialize the distribution, calls
        :mod:`smrf.distribute.image_data.image_data._initialize`. Checks for
        the enhancement factors for the stations and vegetation.

        Args:
            topo: :mod:`smrf.data.loadTopo.Topo` instance contain topographic
                data and infomation
            data: data Pandas dataframe containing the station data,
                from :mod:`smrf.data.loadData` or :mod:`smrf.data.loadGrid`

        """

        self._logger.debug('Initializing distribute.wind')
        self.date_time = date_time
        self.wind_model._initialize(topo, data.metadata)

        if self.check_wind_model_type('winstral'):
            self.add_thread_variables(self.wind_model.thread_variables)

        if not self.check_wind_model_type('interp'):
            self.wind_model.initialize(topo, data)

    def distribute(self, data_speed, data_direction, t):
        """
        Distribute given a Panda's dataframe for a single time step. Calls
        :mod:`smrf.distribute.image_data.image_data._distribute` for
        the `wind_model` chosen.

        Args:
            data_speed: Pandas dataframe for single time step from wind_speed
            data_direction: Pandas dataframe for single time step from
                wind_direction
            t: time stamp

        """

        self._logger.debug('{} Distributing wind_direction and wind_speed'
                           .format(data_speed.name))

        if self.check_wind_model_type('interp'):

            self._distribute(data_speed, other_attribute='wind_speed')

            # wind direction components at the station
            self.u_direction = np.sin(data_direction * np.pi/180)
            self.v_direction = np.cos(data_direction * np.pi/180)

            # distribute u_direction and v_direction
            self._distribute(self.u_direction,
                             other_attribute='u_direction_distributed')
            self._distribute(self.v_direction,
                             other_attribute='v_direction_distributed')

            # combine u and v to azimuth
            az = np.arctan2(self.u_direction_distributed,
                            self.v_direction_distributed)*180/np.pi
            az[az < 0] = az[az < 0] + 360
            self.wind_direction = az

        else:
            self.wind_model.distribute(data_speed, data_direction)

        for v in self.output_variables.keys():
            setattr(self, v, getattr(self.wind_model, v))

        # set min and max
        self.wind_speed = utils.set_min_max(self.wind_speed,
                                            self.wind_model.min,
                                            self.wind_model.max)

    def distribute_thread(self, smrf_queue, data_queue):
        """
        Distribute the data using threading. All data is provided and
        ``distribute_thread`` will go through each time step and call
        :mod:`smrf.distribute.wind.wind.distribute` then puts the distributed
        data into the smrf_queue for :py:attr:`wind_speed`.

        Args:
            smrf_queue: smrf_queue dictionary for all variables
            data: pandas dataframe for all data, indexed by date time
        """
        self._logger.info("Distributing {}".format(self.variable))

        for date_time in self.date_time:

            ws_data = data_queue['wind_speed'].get(date_time)
            wd_data = data_queue['wind_direction'].get(date_time)

            self.distribute(ws_data, wd_data, date_time)

            smrf_queue['wind_speed'].put(
                [date_time, self.wind_model.wind_speed])
            smrf_queue['wind_direction'].put(
                [date_time, self.wind_model.wind_direction])

            if self.check_wind_model_type('winstral'):
                smrf_queue['flatwind'].put(
                    [date_time, self.wind_model.flatwind])
                smrf_queue['cellmaxus'].put(
                    [date_time, self.wind_model.cellmaxus])
                smrf_queue['dir_round_cell'].put(
                    [date_time, self.wind_model.dir_round_cell])



import os

import numpy as np
import pandas as pd
import utm

from smrf.data import Topo
from smrf.envphys import sunang
from smrf.envphys.solar import irradiance, twostream
from smrf.tests.smrf_test_case import SMRFTestCase


class TestRadiation(SMRFTestCase):

    def test_sunang(self):
        """ Sunang calculation """

        # replicate the sun angle calculation from the sunang man page
        # for Santa Barbara on Feb 15, 1990 at 20:30 UTC
        #
        #           | IPW
        # zenith    | 47.122
        # cosz      | 0.680436
        # azimuth   | -5.413
        #
        # The differences between the IPW version and python version
        # are insignificant and are only different because of the
        # values are pulled from stdout for IPW which uses printf

        date_time = pd.to_datetime('2/15/1990 20:30')
        date_time = date_time.tz_localize('UTC')
        lat = 34.4166667  # 35d, 25m, 0s
        lon = -119.9
        ipw_cosz = 0.680436
        ipw_azimuth = -5.413
        ipw_rad_vector = 0.98787

        # try out the python version
        result = sunang.sunang(date_time, lat, lon)
        self.assertTrue(result[0], ipw_cosz)
        self.assertTrue(result[1], ipw_azimuth)
        self.assertTrue(result[2], ipw_rad_vector)

    def test_sunang_functions(self):
        """ Sunang functions """

        self.assertTrue(sunang.leapyear(2016))
        self.assertFalse(sunang.leapyear(2013))

        self.assertEqual(sunang.yearday(2016, 1, 31), 31)
        self.assertEqual(sunang.yearday(2016, 12, 31), 366)
        self.assertEqual(sunang.yearday(2015, 12, 31), 365)
        self.assertEqual(sunang.yearday(2015, 10, 1), 274)
        self.assertEqual(sunang.yearday(2016, 10, 1), 275)

        # replicating the IPW sunang example
        date_time = pd.to_datetime('2/15/1990 20:30')
        date_time = date_time.tz_localize('UTC')
        declin, omega, r = sunang.ephemeris(date_time)
        self.assertEqual(round(declin, 9), -0.218992538)
        self.assertEqual(round(omega, 9), -2.163529935)
        self.assertEqual(round(r, 9), 0.987871247)

    def test_sunang_array(self):
        """ Sunang as a numpy array """

        date_time = pd.to_datetime('2/15/1990 20:30')
        date_time = date_time.tz_localize('UTC')

        topo_config = {
            'filename': os.path.join(self.basin_dir, 'topo/topo.nc'),
            'northern_hemisphere': True,
            'gradient_method': 'gradient_d8',
            'sky_view_factor_angles': 72
        }
        topo = Topo(topo_config)

        # convert from UTM to lat/long
        lat, lon = utm.to_latlon(topo.X[0, 0], topo.Y[0, 0], 11, 'N')
        lat = np.nan * np.zeros_like(topo.X)
        lon = np.nan * np.zeros_like(topo.X)
        for idx, x in np.ndenumerate(topo.X):
            lat[idx], lon[idx] = utm.to_latlon(
                topo.X[idx], topo.Y[idx], 11, 'N')

        self.assertFalse(np.any(np.isnan(lat)))
        self.assertFalse(np.any(np.isnan(lon)))

        cosz, azimuth, rad_vec = sunang.sunang(date_time, lat, lon)

        self.assertTrue(isinstance(cosz, np.ndarray))
        self.assertTrue(isinstance(azimuth, np.ndarray))
        self.assertTrue(isinstance(rad_vec, float))

    def test_direct_solar_irradiance(self):
        """ Test the solar function """

        # output from solar IPW function
        sin = 150.214

        # The example in solar
        date_time = pd.to_datetime('6/22/1990 00:00')
        date_time = date_time.tz_localize('UTC')

        # Python version
        spy = irradiance.direct_solar_irradiance(date_time, w=[0.58, 0.68])
        self.assertTrue(np.abs(spy - sin) <= 0.021)

    def test_twostream(self):
        """ Twostream calculation """

        # IPW command from man twostream is the first test
        # twostream -u 0.68 -t 0.2 -w 0.85 -g 0.3 -r 0.5 -s 159
        # twostream -u 0.05 -t 0.3 -w 0.8 -g 0.35 -r 0.6 -s 60
        # twostream -u 0.88 -t 0.2 -w 0.7 -g 0.3 -r 0.55 -s 1200

        # inputs
        cosz = [0.68, 0.05, 0.88]
        tau = [0.2, 0.3, 0.2]
        omega = [0.85, 0.8, 0.7]
        g = [0.3, 0.35, 0.3]
        R0 = [0.5, 0.6, 0.55]
        S0 = [159, 60, 1200]

        # outputs
        reflectance = [0.472814, 0.577519, 0.465302]
        transmittance = [0.912634, 0.365645, 0.912601]
        direct_transmittance = [0.745189, 0.00247875, 0.796703]
        upwelling_irradiance = [51.1207, 1.73256, 491.359]
        total_irradiance_at_bottom = [98.674, 1.09694, 963.707]
        direct_irradiance_normal_to_beam = [118.485, 0.148725, 956.044]

        for n in [0, 1]:

            # Python way
            py_R = twostream.twostream(
                cosz[n], S0[n], tau=tau[n], omega=omega[n], g=g[n], R0=R0[n])

            # IPW is printed to %g format so convert
            for key, value in py_R.items():
                py_R[key] = float('{:g}'.format(value[0]))
            self.assertTrue(py_R['reflectance'] == reflectance[n])
            self.assertTrue(py_R['transmittance'] == transmittance[n])
            self.assertTrue(py_R['direct_transmittance']
                            == direct_transmittance[n])
            self.assertTrue(py_R['upwelling_irradiance']
                            == upwelling_irradiance[n])
            self.assertTrue(py_R['irradiance_at_bottom']
                            == total_irradiance_at_bottom[n])
            self.assertTrue(py_R['irradiance_normal_to_beam']
                            == direct_irradiance_normal_to_beam[n])



from inicheck.tools import cast_all_variables


class BaseRecipes():
    """Test the recipes
    """

    def cast_recipes(self, config):
        """Cast the inicheck recipes
        Arguments:
            config {UserConfig} -- UserConfig object to modify
        Returns:
            UserConfig -- Modified UserConfig object
        """

        config.apply_recipes()
        config = cast_all_variables(config, config.mcfg)

        return config

    def master_config(self, config):
        """Create a master config dictionary with a
        list of the keys
        Arguments:
            config {UserConfig} -- UserConfig object to create
                a master config from
        Returns:
            [dict] -- master_config dict that has all sections
                and items from the provided UserConfig
        """

        master_config = {}
        for key in config.mcfg.cfg.keys():
            master_config[key] = list(config.mcfg.cfg[key].keys())

        return master_config

    def check_config(self, config, master_config):
        """Check that the config read by inicheck matches
        the master_config (expected) results
        Arguments:
            config {inicheck UserConfig} -- UserConfig object to check
            master_config {dict} -- dict of sections and items
                that should be in the UserConfig
        """

        # ensure that there are no
        for key, items in master_config.items():
            self.check_config_section(config, master_config, key)

    def check_config_section(self, config, master_config, section):
        """Check that the config read by inicheck matches
        the master_config (expected) results
        Arguments:
            config {inicheck UserConfig} -- UserConfig object to check
            master_config {dict} -- dict of sections and items
                that should be in the UserConfig
        """

        # ensure that there are no
        items = master_config[section]
        test_items = list(config.cfg[section].keys())

        for item in items:
            self.assertTrue(item in test_items)
            test_items.remove(item)

        self.assertTrue(len(test_items) == 0)



"""
Functions to output the gridded data for a HRU
"""

import logging
import os

import numpy as np
import pandas as pd


class output_hru():
    """
    Class output_hru() to output values to a HRU dataframe, then to a file
    """

    fmt = '%Y-%m-%d %H:%M:%S'
    date_cols = ["year", "month", "day", "hour", "minute", "second"]

    def __init__(self, variable_list, topo, date_time, config):
        """
        Initialize the output_hru() class

        Args:
            variable_list: list of dicts, one for each variable
            topo: loadTopo instance
            date_time: smrf.date_time, array of dates for the SMRF run
            config: configuration for the output
        """

        self._logger = logging.getLogger(__name__)
        self.config = config

        # check some defaults based on the output type
        self.prms_flag = False
        if config['output_type'] == 'csv':
            ext = '.csv'
            self.delimiter = ','

        elif config['output_type'] == 'prms':
            ext = '.data'
            self.prms_flag = True
            if not os.path.isfile(self.config['hru_file']):
                raise Exception('HRU file does not exist')
            self.delimiter = ' '

        # go through the variable list and make full file names
        for v in variable_list:
            if not variable_list[v]['out_location'].endswith(ext):
                variable_list[v]['file_name'] = \
                    variable_list[v]['out_location'] + ext
            else:
                variable_list[v]['file_name'] = \
                    variable_list[v]['out_location']

            try:
                os.remove(variable_list[v]['file_name'])
            except OSError:
                pass

        self.variable_list = variable_list

        # process the time section
        self.out_frequency = int(config['frequency'])
        self.date_time = date_time
        self.idx = 0

        # read in the HRU file
        self._logger.debug('Reading HRU ascii {}'
                           .format(self.config['hru_file']))

        hru = np.loadtxt(self.config['hru_file'], skiprows=6)
        hru_max = int(np.max(hru))
        self._logger.debug("Number of HRU's: {}".format(hru_max))

        self.hru = hru
        self.hru_max = hru_max

        self.IND = {}
        for h in range(hru_max):
            self.IND[h] = hru == h+1

        self.hru_idx = [str(i+1) for i in range(self.hru_max)]

        # create an empty dataframe
        if self.prms_flag:
            cols = self.date_cols + self.hru_idx
            hru_data = pd.DataFrame(index=range(len(self.date_time)),
                                    columns=cols)

            yrs = np.array([
                [y.year, y.month, y.day, y.hour, y.minute, y.second]
                for y in self.date_time])
            hru_data['year'] = yrs[:, 0]
            hru_data['month'] = yrs[:, 1]
            hru_data['day'] = yrs[:, 2]
            hru_data['hour'] = yrs[:, 3]
            hru_data['minute'] = yrs[:, 4]
            hru_data['second'] = yrs[:, 5]

            hru_data.loc[:, self.hru_idx] = \
                hru_data.loc[:, self.hru_idx].apply(pd.to_numeric)

            # lets start the output file
            self.generate_prms_header()

        else:
            cols = ["date_time"] + self.hru_idx
            hru_data = pd.DataFrame(index=range(len(self.date_time)),
                                    columns=cols)
            hru_data['date_time'] = self.date_time

        self.hru_data = hru_data

#     def generate_csv_header(self, hru_data):
#         """
#         Generate the header for the csv file
#         """
#
#         for v in self.variable_list:
#
#             with open(self.variable_list[v]['file_name'],'w') as f:
#
#
#                 hdr.to_csv(f, self.delimiter, index=False, header=True)

    def generate_prms_header(self):
        """
        Generate the header for the PRMS output file
        """

        for v in self.variable_list:

            with open(self.variable_list[v]['file_name'], 'w') as f:
                # Append the PRMS expected Header
                f.write("File Generated using SMRF distributed forcing data\n")
                f.write("{0} {1}\n".format(v, self.hru_max))

                f.write("########################################\n")
                f.close()

    def output(self, variable, data, date_time):
        """
        Output a time step

        Args:
            variable: variable name that will index into variable list
            data: the variable data
            date_time: the date time object for the time step
        """

        self._logger.debug('{} Writing variable {} to {} file'
                           .format(
                               date_time,
                               variable,
                               self.config['output_type']))

        # loop through the HRU
        m_hru = np.zeros((self.hru_max,))
        for h in range(self.hru_max):
            m_hru[h] = np.nanmean(data[self.IND[h]])

#             self.hru_data.loc[self.idx, (str(h+1))] = m_hru[h]

        if self.func == 'mm2in':
            m_hru /= 25.4
        elif self.func == 'C2F':
            m_hru = m_hru * 9/5 + 32

        self.hru_data.loc[self.idx, self.hru_idx] = m_hru

        # open the file and write the row of data
        with open(self.variable_list[variable]['file_name'], 'a') as f:

            row = self.hru_data.iloc[self.idx].to_frame().T

            # change the dates columns back to int
            if self.prms_flag:
                row[self.date_cols] = row[self.date_cols].astype(int)

            hdr = None
            if (self.idx == 0) & (not self.prms_flag):
                hdr = True

            row.to_csv(f, sep=self.delimiter,
                       header=hdr,
                       index=False,
                       float_format='%.3f')

            f.close()

        self.idx += 1



from smrf.distribute import image_data
from smrf.utils import utils


class cf(image_data.image_data):
    """
    The :mod:`~smrf.distribute.cloud_factor.cf` class allows for variable
    specific distributions that go beyond the base class. Cloud factor is
    a relatively simple variable to distribute as it does not rely on any
    other variables.

    Cloud factor is calculated as the ratio between measured incoming
    solar radiation and modeled clear sky radiation. A value of 0 means
    no incoming solar radiation (or very cloudy) and a value of 1 means
    sunny.

    Args:
        config: The [cloud_factor] section of the configuration file

    Attributes:
        config: configuration from [cloud_factor] section
        cloud_factor: numpy array of the cloud factor
        stations: stations to be used in alphabetical order

    """

    variable = 'cloud_factor'

    # these are variables that can be output
    OUTPUT_VARIABLES = {
        'cloud_factor': {
            'units': 'None',
            'standard_name': 'cloud_factor',
            'long_name': 'cloud factor'
        }
    }

    # these are variables that are operate at the end only and do not need to
    # be written during main distribute loop
    post_process_variables = {}

    BASE_THREAD_VARIABLES = frozenset([
        'cloud_factor'
    ])

    def __init__(self, config):

        # extend the base class
        image_data.image_data.__init__(self, self.variable)

        # check and assign the configuration
        self.getConfig(config)
        self._logger.debug('Created distribute.cloud_factor')

    def initialize(self, topo, data, date_time=None):
        """
        Initialize the distribution, solely calls
        :mod:`smrf.distribute.image_data.image_data._initialize`.

        Args:
            topo: :mod:`smrf.data.loadTopo.Topo` instance contain topographic
                data and infomation
            metadata: metadata Pandas dataframe containing the station metadata
                from :mod:`smrf.data.loadData` or :mod:`smrf.data.loadGrid`

        """

        self._logger.debug('Initializing distribute.cloud_factor')
        self.date_time = date_time
        self._initialize(topo, data.metadata)

    def distribute(self, data):
        """
        Distribute cloud factor given a Panda's dataframe for a single time
        step. Calls :mod:`smrf.distribute.image_data.image_data._distribute`.

        Args:
            data: Pandas dataframe for a single time step from cloud_factor

        """

        self._logger.debug('{} Distributing cloud_factor'.format(data.name))

        self._distribute(data)
        self.cloud_factor = utils.set_min_max(self.cloud_factor, self.min,
                                              self.max)

    def distribute_thread(self, smrf_queue, data_queue):
        """
        Distribute the data using threading and queue. All data is provided
        and ``distribute_thread`` will go through each time step and call
        :mod:`smrf.distribute.cloud_factor.cf.distribute` then puts the
        distributed data into ``queue['cloud_factor']``.

        Args:
            queue: queue dictionary for all variables
            data: pandas dataframe for all data, indexed by date time

        """
        self._logger.info("Distributing {}".format(self.variable))

        for date_time in self.date_time:

            data = data_queue[self.variable].get(date_time)
            self.distribute(data)
            smrf_queue[self.variable].put([date_time, self.cloud_factor])



import netCDF4 as nc
import numpy as np
import pytz

from smrf.data import Topo
from smrf.distribute.wind.wind_ninja import WindNinjaModel
from smrf.tests.smrf_test_case_lakes import SMRFTestCaseLakes
from smrf.utils import utils
from smrf.utils.utils import date_range


class TestWindNinja(SMRFTestCaseLakes):

    def test_wind_ninja(self):

        config = self.base_config.cfg

        # Load the topo
        topo = Topo(config['topo'])

        # Get the time steps correctly in the time zone
        tzinfo = pytz.timezone(config['time']['time_zone'])
        date_time = date_range(
            config['time']['start_date'],
            config['time']['end_date'],
            config['time']['time_step'],
            tzinfo)

        wn = WindNinjaModel(config)
        wn.initialize(topo, None)
        wn.initialize_interp(date_time[0])
        g_vel, g_ang = wn.convert_wind_ninja(date_time[0])

        # The x values are ascending
        self.assertTrue(np.all(np.diff(wn.windninja_x) > 0))

        # The y values are descnding
        self.assertTrue(np.all(np.diff(wn.windninja_y) < 0))

        # compare against gold
        g_vel = utils.set_min_max(
            g_vel, wn.config['min'], wn.config['max'])

        # check against gold
        # The two are not exactly the same as there is some float
        # precision error with netcdf
        n = nc.Dataset(self.gold_dir.joinpath('wind_speed.nc'))
        np.testing.assert_allclose(
            n.variables['wind_speed'][0, :],
            g_vel
        )
        n.close()

        n = nc.Dataset(self.gold_dir.joinpath('wind_direction.nc'))
        np.testing.assert_allclose(
            n.variables['wind_direction'][0, :],
            g_ang
        )
        n.close()



import numpy as np

from smrf.envphys.constants import FREEZE, STEF_BOLTZ
from smrf.envphys.thermal.clear_sky import calc_long_wave


def Garen2005(th, cloud_factor):
    """
    Cloud correction is based on the relationship in Garen and Marks (2005)
    :cite:`Garen&Marks:2005` between the cloud factor and measured long
    wave radiation using measurement stations in the Boise River Basin.

    .. math::
        L_{cloud} = L_{clear} * (1.485 - 0.488 * cloud\\_factor)

    Args:
        th: clear sky thermal radiation [W/m2]
        cloud_factor: fraction of sky that are not clouds, 1 equals no clouds,
            0 equals all clouds

    Returns:
        cloud corrected clear sky thermal radiation

    20170515 Scott Havens
    """

    return th * (1.485 - 0.488 * cloud_factor)


def Unsworth1975(th, ta, cloud_factor):
    """
    Cloud correction is based on Unsworth and Monteith (1975)
    :cite:`Unsworth&Monteith:1975`

    .. math::
            \\epsilon_a = (1 - 0.84) \\epsilon_{clear} + 0.84c

    where :math:`c = 1 - cloud\\_factor`

    Args:
        th: clear sky thermal radiation [W/m2]
        ta: temperature in Celcius that the clear sky thermal radiation was
            calcualted from [C] cloud_factor: fraction of sky that are not
            clouds, 1 equals no clouds, 0 equals all clouds

    Returns:
        cloud corrected clear sky thermal radiation

    20170515 Scott Havens
    """

    c = 1 - cloud_factor
    ta = ta + FREEZE

    # calculate the clear sky emissivity
    ec = th/(STEF_BOLTZ * np.power(ta, 4))

    # adjsut the atmospheric emissivity for clouds
    ea = (1 - 0.84*c) * ec + 0.84*c

    return calc_long_wave(ea, ta)


def Kimball1982(th, ta, ea, cloud_factor):
    """
    Cloud correction is based on Kimball et al. (1982) :cite:`Kimball&al:1982`

    .. math::

        L_d &= L_{clear} + \\tau_8 c f_8 \\sigma T^{4}_{c}

        \\tau_8 &= 1 - \\epsilon_{8z} (1.4 - 0.4 \\epsilon_{8z})

        \\epsilon_{8z} &= 0.24 + 2.98 \\times 10^{-6} e^2_o exp(3000/T_o)

        f_8 &= -0.6732 + 0.6240 \\times 10^{-2} T_c - 0.9140
        \\times 10^{-5} T^2_c

    where the original Kimball et al. (1982) :cite:`Kimball&al:1982` was for
    multiple cloud layers, which was simplified to one layer. :math:`T_c` is
    the cloud temperature and is assumed to be 11 K cooler than :math:`T_a`.

    Args:
        th: clear sky thermal radiation [W/m2]
        ta: temperature in Celcius that the clear sky thermal radiation was
            calcualted from [C]
        ea: distrubted vapor pressure [kPa]
        cloud_factor: fraction of sky that are not clouds, 1 equals no clouds,
            0 equals all clouds

    Returns:
        cloud corrected clear sky thermal radiation

    20170515 Scott Havens
    """

    ta = ta + FREEZE
    Tc = ta - 11

    f8 = -0.6732 + 0.6240 * 10**(-2) * Tc - 0.9140 * 10**(-5) * Tc**2
    e8z = 0.24 + 2.98 * 10**(-6) * ea**2 * np.exp(3000/ta)
    t8 = 1 - e8z * (1.4 - 0.4 * e8z)

    return th + calc_long_wave(f8 * e8z * t8, ta)


def Crawford1999(th, ta, cloud_factor):
    """
    Cloud correction is based on Crawford and Duchon (1999)
    :cite:`Crawford&Duchon:1999`

    .. math::
            \\epsilon_a = (1 - cloud\\_factor) +
            cloud\\_factor * \\epsilon_{clear}

    where :math:`cloud\\_factor` is the ratio of measured solar radiation
    to the clear sky irradiance.

    Args:
        th: clear sky thermal radiation [W/m2]
        ta: temperature in Celcius that the clear sky thermal radiation was
            calcualted from [C]
        cloud_factor: fraction of sky that are not clouds, 1 equals no clouds,
            0 equals all clouds

    Returns:
        cloud corrected clear sky thermal radiation

    20170515 Scott Havens
    """

    return calc_long_wave(1 - cloud_factor, ta + FREEZE) + cloud_factor * th



import pandas as pd

from smrf.envphys.solar.model import model_solar


def cf_cloud(beam, diffuse, cf):
    """
    Correct beam and diffuse irradiance for cloud attenuation at a single
    time, using input clear-sky global and diffuse radiation calculations
    supplied by locally modified toporad or locally modified stoporad

    Args:
        beam: global irradiance
        diffuse: diffuse irradiance
        cf: cloud attenuation factor - actual irradiance / clear-sky irradiance

    Returns:
        c_grad: cloud corrected gobal irradiance
        c_drad: cloud corrected diffuse irradiance

    20150610 Scott Havens - adapted from cloudcalc.c
    """

    # define some constants
    CRAT1 = 0.15
    CRAT2 = 0.99
    CCOEF = 1.38

    # cloud attenuation, beam ratio is reduced
    bf_c = CCOEF * (cf - CRAT1)**2
    c_grad = beam * cf
    c_brad = c_grad * bf_c
    c_drad = c_grad - c_brad

    # extensive cloud attenuation, no beam
    ind = cf <= CRAT1
    c_brad[ind] = 0
    c_drad[ind] = c_grad[ind]

    # minimal cloud attenution, no beam ratio reduction
    ind = cf > CRAT2
    c_drad[ind] = diffuse[ind] * cf[ind]
    c_brad[ind] = c_grad[ind] - c_drad[ind]

    return c_grad, c_drad


def get_hrrr_cloud(df_solar, df_meta, lat, lon):
    """
    Take the combined solar from HRRR and use the two stream calculation
    at the specific HRRR pixels to find the cloud_factor.

    Args:
        df_solar - solar dataframe from hrrr
        df_meta - meta_data from hrrr
        lat - basin lat
        lon - basin lon

    Returns:
        df_cf - cloud factor dataframe in same format as df_solar input
    """

    # get and loop through the columns
    dates = df_solar.index.values[:]

    # find each cell solar at top of atmosphere for each date
    basin_sol = df_solar.copy()
    for idt, dt in enumerate(dates):
        # get solar using twostream
        dtt = pd.to_datetime(dt)
        basin_sol.iloc[idt, :] = model_solar(dtt, lat, lon)

    # if it's close to sun down or sun up, then the cloud factor gets
    # difficult to calculate
    basin_sol[basin_sol < 50] = 0
    df_solar[basin_sol < 50] = 0

    # This will produce NaN values when the sun is down
    df_cf = df_solar / basin_sol

    # linear interpolate the NaN values at night
    df_cf = df_cf.interpolate(method='linear').ffill()

    # Clean up the dataframe to be between 0 and 1
    df_cf[df_cf > 1.0] = 1.0
    df_cf[df_cf < 0.0] = 0.0

    return df_cf



import numpy as np
import pandas as pd


def time_since_storm(precipitation, perc_snow, time_step=1/24, mass=1.0,
                     time=4, stormDays=None, stormPrecip=None, ps_thresh=0.5):
    """
    Calculate the decimal days since the last storm given a precip time series,
    percent snow, mass threshold, and time threshold

     - Will look for pixels where perc_snow > 50% as storm locations
     - A new storm will start if the mass at the pixel has exceeded the mass
         limit, this ensures that the enough has accumulated

    Args:
        precipitation: Precipitation values
        perc_snow: Percent of precipitation that was snow
        time_step: Step in days of the model run
        mass: Threshold for the mass to start a new storm
        time: Threshold for the time to start a new storm
        stormDays: If specified, this is the output from a previous run of
            storms else it will be set to the date_time value
        stormPrecip: Keeps track of the total storm precip

    Returns:
        tuple:
        - **stormDays** - Array representing the days since the last storm at
            a pixel
        - **stormPrecip** - Array representing the precip accumulated during
            the most recent storm

    Created Janurary 5, 2016
    @author: Scott Havens
    """
    # either preallocate or use the input
    if stormDays is None:
        stormDays = np.zeros(precipitation.shape)

    if stormPrecip is None:
        stormPrecip = np.zeros(precipitation.shape)

    # if there is no snow, don't reset the counter
    # This ensures that the albedo won't be reset
    stormDays += 1
    if np.sum(perc_snow) == 0:
        #         stormDays = np.add(stormDays, 1)
        stormPrecip = np.zeros(precipitation.shape)
        return stormDays, stormPrecip

    # determine locations where it has snowed
    idx = perc_snow >= ps_thresh

    # determine locations where the time threshold has passed
    # these areas, the stormPrecip will be set back to zero
    idx_time = stormDays >= time
    stormPrecip[idx_time] = 0

    # add the values to the stormPrecip
    stormPrecip[idx] = + precipitation[idx]

    # see if the mass threshold has been passed
    idx_mass = stormPrecip >= mass

    # reset the stormDays to zero where the storm is present
    stormDays[idx_mass] = 0

    return stormDays, stormPrecip


def time_since_storm_pixel(precipitation, dpt, perc_snow, storming,
                           time_step=1/24, stormDays=None, mass=1.0,
                           ps_thresh=0.5):
    """
    Calculate the decimal days since the last storm given a precip time series

     - Will assign decimal days since last storm to every pixel

    Args:
        precipitation: Precipitation values
        dpt: dew point values
        perc_snow: percent_snow values
        storming: if it is stomring
        time_step: step in days of the model run
        stormDays: unifrom days since last storm on pixel basis
        mass: Threshold for the mass to start a new storm
        ps_thresh: Threshold for percent_snow

    Returns:
        stormDays: days since last storm on pixel basis

    Created October 16, 2017
    @author: Micah Sandusky
    """
    # either preallocate or use the input
    if stormDays is None:
        stormDays = np.zeros(precipitation.shape)

    # add timestep
    stormDays += time_step

    # only reset if stomring and not overly warm
    if storming and dpt.min() < 2.0:
        # determine location where there is enough mass
        idx_mass = precipitation >= mass
        # determine locations where it has snowed
        idx = perc_snow >= ps_thresh

        # reset the stormDays to zero where the storm is present
        stormDays[(idx_mass & idx)] = 0

    return stormDays


def tracking_by_station(precip, mass_thresh=0.01, steps_thresh=3):
    """
    Processes the vector station data prior to the data being distributed

    Args:
        precipitation: precipitation values
        time: Time step that smrf is on
        time_steps_since_precip: time steps since the last precipitation
        storm_lst: list that store the storm cycles in order. A storm is
                    recorded by its start and its end. The list
                    is passed by reference and modified internally.
                    Each storm entry should be in the format of:
                    [{start:Storm Start, end:Storm End}]

                    e.g.
                    [
                    {start:date_time1,end:date_time2,'BOG1':100, 'ATL1':85},
                    {start:date_time3,end:date_time4,'BOG1':50, 'ATL1':45},
                    ]

                    would be a two storms at stations BOG1 and ATL1

        mass_thresh: mass amount that constitutes a real precip event,
            default = 0.01.

        steps_thresh: Number of time steps that constitutes the end of a precip
            event, default = 2 steps (typically 2 hours)

    Returns:
        tuple:
            - **storms** - A list of dictionaries containing storm start,stop,
                mass accumulated, of given storm.

            - **storm_count** - A total number of storms found

    Created April 24, 2017
    @author: Micah Johnson
    """

    storm_columns = ['start', 'end']
    stations = list(precip)
    storm_columns += stations

    storms = []

    stations = list(precip)
    is_storming = False
    time_steps_since_precip = 0

    for i, row in precip.iterrows():
        time = pd.Timestamp(i)

        # Storm Idenificiation
        if row.max() > mass_thresh:
            # Start a new storm
            if not is_storming:
                new_storm = {}
                new_storm['start'] = time
                for sta, p in row.iteritems():
                    new_storm[sta] = 0
                # Create a new row
                is_storming = True

            time_steps_since_precip = 0
            # Always add the latest end date to avoid unclosed storms
            new_storm['end'] = time

            # Accumulate precip for storm total
            for sta, mass in row.iteritems():
                new_storm[sta] += mass

        elif is_storming and time_steps_since_precip < steps_thresh:
            new_storm['end'] = time

            time_steps_since_precip += 1

        if time_steps_since_precip >= steps_thresh and is_storming:
            is_storming = False
            storms.append(new_storm)
            # print "=="*10 + "> not storming!"

    # Append the last storm if we ended during a storm
    if is_storming:
        storms.append(new_storm)

    storm_count = len(storms)

    # Make sure we have storms
    if storm_count == 0:
        empty_data = {}
        for col in storm_columns:
            empty_data[col] = []
        storms = pd.DataFrame(empty_data)
    else:
        storms = pd.DataFrame(storms)

    return storms, storm_count


def tracking_by_basin(precipitation, time, storm_lst, time_steps_since_precip,
                      is_storming, mass_thresh=0.01, steps_thresh=2):
    """
    Args:
        precipitation: precipitation values
        time: Time step that smrf is on
        time_steps_since_precip: time steps since the last precipitation
        storm_lst: list that store the storm cycles in order. A storm is
                    recorded by its start and its end. The list
                    is passed by reference and modified internally.
                    Each storm entry should be in the format of:
                    [{start:Storm Start, end:Storm End}]

                    e.g.
                         [
                         {start:date_time1,end:date_time2},
                         {start:date_time3,end:date_time4},
                         ]

                         #would be a two storms

        mass_thresh: mass amount that constitutes a real precip
                    event, default = 0.0.
        steps_thresh: Number of time steps that constitutes the end of
                        a precip event, default = 2 steps (default 2 hours)

    Returns:
        tuple:
            storm_lst - updated storm_lst
            time_steps_since_precip - updated time_steps_since_precip
            is_storming - True or False whether the storm is ongoing or not

    Created March 3, 2017
    @author: Micah Johnson
    """
    # print  "--"*10 +"> Max precip = {0}".format(precipitation.max())
    if precipitation.max() > mass_thresh:
        # Start a new storm
        if len(storm_lst) == 0 or not is_storming:
            storm_lst.append({'start': time, 'end': None})
            is_storming = True

        # always append the most recent timestep to avoid unended storms
        storm_lst[-1]['end'] = time
        time_steps_since_precip = 0

    elif is_storming and time_steps_since_precip < steps_thresh:
        time_steps_since_precip += 1

    if time_steps_since_precip >= steps_thresh:
        is_storming = False
        # print "--"*10 + "> not storming!"

    return storm_lst, time_steps_since_precip, is_storming


def clip_and_correct(precip, storms, stations=[]):
    """
    Meant to go along with the storm tracking, we correct the data here by
    adding in the precip we would miss by ignoring it. This is mostly because
    will get rain on snow events when there is snow because of the storm
    definitions and still try to distribute precip data.

    Args:
        precip: Vector station data representing the measured precipitation
        storms: Storm list with dictionaries as defined in
                :func:`~smrf.envphys.storms.tracking_by_station`
        stations: Desired stations that are being used for clipping. If
                  stations is not passed, then use all in the dataframe


    Returns:
        The correct precip that ensures there is no precip outside of the
        defined storms with the clipped amount of precip proportionally added
        back to storms.

    Created May 3, 2017
    @author: Micah Johnson
    """

    # Specify zeros where were not storming
    precip_clipped = precip.copy()
    precip_clipped[:] = 0

    for j, storm in storms.iterrows():

        storm_start = storm['start']
        storm_end = storm['end']
        my_slice = precip.loc[storm_start:storm_end]
        precip_clipped.loc[storm_start:storm_end] = my_slice

    correction = {}

    if len(stations) == 0:
        stations = precip.columns

    # Correct the precip to be equal to the sum.
    for station in stations:
        original = precip[station].sum()
        clipped = precip_clipped[station].sum()

        if original == 0:
            c = 1.0
        elif clipped == 0:
            c = 0
        else:
            c = original/clipped

        correction[station] = c

    return precip_clipped.mul(pd.Series(correction), axis=1)



'''
Created on Apr 15, 2015

@author: scott
'''

import numpy as np

from smrf.utils import utils


def catchment_ratios(ws, gauge_type, snowing):
    """
    Point models for catchment ratios of the
    """

    if gauge_type == "us_nws_8_shielded":
        if snowing:
            CR = np.exp(4.61 - 0.04*ws**1.75)
        else:
            CR = 101.04 - 5.62*ws

    elif gauge_type == "us_nws_8_unshielded":
        if snowing:
            CR = np.exp(4.61 - 0.16*ws**1.28)
        else:
            CR = 100.77 - 8.34*ws
    else:
        raise ValueError(
            "Unknown catchement adjustment model ----> {0}".format(gauge_type))

    #    elif type == "Hellmann unshielded":
    #        if snowing:
    #             CR = CR = 100.00 + 1.13*Ws - 19.45*Ws
    #        else:
    #             CR = 96.63 + 0.41*Ws - 9.84*Ws + 5.95 * Tmean

    # elif type == "nipher":
    #     if snowing:
    #          CR= 100.00 - 0.44*Ws - 1.98*Ws
    #     else:
    #          CR = 97.29 - 3.18*Ws+ 0.58* Tmax - 0.67*Tmin
    #
    # elif type == "tretyakov":
    #     if snowing:
    #          CR = 103.11 - 8.67 * Ws + 0.30 * Tmax
    #     else:
    #          CR =  96.99 - 4.46 *Ws + 0.88 * Tmax + 0.22*Tmin
    # Avoid corrupting data
    if np.isnan(CR):
        CR = 100.0
    # CR is in percent.
    CR = CR/100.0
    return CR


def adjust_for_undercatch(p_vec, wind, temp, sta_type, metadata):
    """
    Adjusts the vector precip station data for undercatchment. Relationships
    should be added to :func:`~smrf.envphys.precip.catchment_ratio`.

    Args:
        p_vec - The station vector data in pandas series
        wind -  The vector wind data
        temp - The vector air_temp data
        sta_type - A dictionary of station names and the type of correction
            to apply
        station_metadata - station metadata TODO merge in the station_dict
            info to metadata

    Returns:
        adj_precip - Adjust precip accoding to the corrections applied.
    """
    adj_precip = p_vec.copy()
    for sta in p_vec.index:
        if sta in temp.keys():
            T = temp[sta]
            if sta in wind.keys():
                ws = wind[sta]
                if ws > 6.0:
                    ws = 6.0

                if T < -0.5:
                    snowing = True
                else:
                    snowing = False

                if sta in sta_type.keys():
                    gauge_type = sta_type[sta]
                else:
                    gauge_type = sta_type['station_undercatch_model_default']

                cr = catchment_ratios(ws, gauge_type, snowing)
                adj_precip[sta] = p_vec[sta]/cr

    return adj_precip


def dist_precip_wind(precip, precip_temp, az, dir_round_cell, wind_speed,
                     cell_maxus, tbreak, tbreak_direction, veg_type, veg_fact,
                     cfg, mask=None):
    """
    Redistribute the precip based on wind speed and direciton
    to account for drifting.

    Args:
        precip:             distributed precip
        precip_temp:        precip_temp array
        az:                 wind direction
        dir_round_cell:     from wind module
        wind_speed:         wind speed array
        cell_maxus:         max upwind slope from maxus file
        tbreak:             relative local slope from tbreak file
        tbreak_direction:   direction array from tbreak file
        veg_type:           user input veg types to correct
        veg_factor:         interception correction for veg types. ppt mult is
                            multiplied by 1/veg_factor

    Returns:
        precip_drift:       numpy array of precip redistributed for wind

    """
    # thresholds
    tbreak_threshold = cfg['tbreak_threshold']
    min_scour = cfg['winstral_min_scour']
    max_scour = cfg['winstral_max_scour']
    min_drift = cfg['winstral_min_drift']
    max_drift = cfg['winstral_max_drift']
    precip_temp_threshold = 0.5

    # polynomial factors
    drift_poly = {}
    drift_poly['a'] = 0.0289
    drift_poly['b'] = -0.0956
    drift_poly['c'] = 1.000761
    ppt_poly = {}
    ppt_poly['a'] = 0.0001737
    ppt_poly['b'] = 0.002549
    ppt_poly['c'] = 0.03265
    ppt_poly['d'] = 0.5929

    # initialize arrays
    celltbreak = np.ones(dir_round_cell.shape)
    drift_factor = np.ones(dir_round_cell.shape)
    precip_drift = np.zeros(dir_round_cell.shape)
    pptmult = np.ones(dir_round_cell.shape)

    # classify tbreak
    dir_unique = np.unique(dir_round_cell)

    for d in dir_unique:
        # find all values for matching direction
        ind = dir_round_cell == d
        i = np.argwhere(tbreak_direction == d)[0][0]
        celltbreak[ind] = tbreak[i][ind]

    # ################################################### #
    # Routine for drift cells
    # ################################################### #
    # for tbreak cells >  threshold
    idx = ((celltbreak > tbreak_threshold) & (
        precip_temp < precip_temp_threshold))

    # calculate drift factor
    drift_factor[idx] = drift_poly['c'] * np.exp(
        drift_poly['b'] * wind_speed[idx] +
        drift_poly['a'] * wind_speed[idx]**2
    )

    # cap drift factor
    drift_factor[idx] = utils.set_min_max(
        drift_factor[idx], min_drift, max_drift)

    # multiply precip by drift factor for drift cells
    precip_drift[idx] = drift_factor[idx]*precip[idx]

    # ################################################### #
    # Now lets deal with non drift cells
    # ################################################### #

    # for tbreak cells <= threshold (i.e. the rest of them)
    idx = ((celltbreak <= tbreak_threshold) & (
        precip_temp < precip_temp_threshold))

    # reset pptmult for exposed pixels
    pptmult = np.ones(dir_round_cell.shape)

    # original from manuscript
    pptmult[idx] = ppt_poly['a'] + ppt_poly['c'] * cell_maxus[idx] - \
        ppt_poly['b'] * cell_maxus[idx]**2 + ppt_poly['a'] * cell_maxus[idx]**3

    # veg effects at indices that we are working on where veg type matches
    for i, v in enumerate(veg_fact):
        idv = ((veg_type == int(v)) & (idx))
        pptmult[idv] = pptmult[idv] * 1.0 / veg_fact[v]

    # hardcode for pine
    pptmult[(idx) & ((veg_type == 3055) | (veg_type == 3053)
                     | (veg_type == 42))] = 1.0  # 0.92
    # cap ppt mult
    pptmult[idx] = utils.set_min_max(pptmult[idx], min_scour, max_scour)

    # multiply precip by scour factor
    precip_drift[idx] = pptmult[idx] * precip[idx]

    # ############################## #
    # no precip redistribution where dew point >= threshold
    idx = precip_temp >= precip_temp_threshold
    precip_drift[idx] = precip[idx]

    return precip_drift



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import unittest

from inicheck.tools import cast_all_variables

from smrf.framework.model_framework import run_smrf
from smrf.tests.smrf_test_case import SMRFTestCase


class TestThreadedRME(SMRFTestCase):
    """
    Integration test for SMRF.
    Runs the short simulation over reynolds mountain east.
    """
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.gold_dir = cls.basin_dir.joinpath('gold')

        run_smrf(cls.config_file)

    def test_air_temp(self):
        self.compare_netcdf_files('air_temp.nc')

    def test_precip_temp(self):
        self.compare_netcdf_files('precip_temp.nc')

    def test_net_solar(self):
        self.compare_netcdf_files('net_solar.nc')

    def test_percent_snow(self):
        self.compare_netcdf_files('percent_snow.nc')

    def test_precip(self):
        self.compare_netcdf_files('precip.nc')

    def test_thermal(self):
        self.compare_netcdf_files('thermal.nc')

    def test_wind_speed(self):
        self.compare_netcdf_files('wind_speed.nc')

    def test_wind_direction(self):
        self.compare_netcdf_files('wind_direction.nc')

    def test_snow_density(self):
        self.compare_netcdf_files('snow_density.nc')

    def test_vapor_pressure(self):
        self.compare_netcdf_files('vapor_pressure.nc')


class TestRME(SMRFTestCase):
    """
    Integration test for SMRF using without threading
    Runs the short simulation over reynolds mountain east
    """
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.gold_dir = cls.basin_dir.joinpath('gold')

        config = cls.base_config_copy()
        config.raw_cfg['system']['threading'] = False

        config.apply_recipes()
        config = cast_all_variables(config, config.mcfg)

        run_smrf(config)

    def test_air_temp(self):
        self.compare_netcdf_files('air_temp.nc')

    def test_precip_temp(self):
        self.compare_netcdf_files('precip_temp.nc')

    def test_net_solar(self):
        self.compare_netcdf_files('net_solar.nc')

    def test_percent_snow(self):
        self.compare_netcdf_files('percent_snow.nc')

    def test_precip(self):
        self.compare_netcdf_files('precip.nc')

    def test_thermal(self):
        self.compare_netcdf_files('thermal.nc')

    def test_wind_speed(self):
        self.compare_netcdf_files('wind_speed.nc')

    def test_wind_direction(self):
        self.compare_netcdf_files('wind_direction.nc')

    def test_snow_density(self):
        self.compare_netcdf_files('snow_density.nc')

    def test_vapor_pressure(self):
        self.compare_netcdf_files('vapor_pressure.nc')


if __name__ == '__main__':
    unittest.main()



import unittest
from io import StringIO
from unittest.mock import patch

from .script_test_helper import load_script_as_module

script = load_script_as_module('gen_maxus')


class TestGenMaxusArguments(unittest.TestCase):
    def setUp(self):
        self.argument_parser = script.argument_parser()

    @patch('sys.stderr', new=StringIO())
    def test_require_dem(self):
        with self.assertRaises(SystemExit):
            self.argument_parser.parse_args([])

    def test_only_requires_dem(self):
        self.argument_parser.parse_args(['dem'])

    def test_output_file_argument(self):
        args = self.argument_parser.parse_args(['dem'])
        self.assertEqual('./maxus.nc', args.out_maxus)

        args = self.argument_parser.parse_args(['dem', '--out_maxus', 'file'])
        self.assertEqual('file', args.out_maxus)

        args = self.argument_parser.parse_args(['dem', '-O', 'file2'])
        self.assertEqual('file2', args.out_maxus)

    def test_increment_argument(self):
        args = self.argument_parser.parse_args(['dem'])
        self.assertEqual(5, args.increment)

        args = self.argument_parser.parse_args(['dem', '--increment', '1'])
        self.assertEqual(1, args.increment)

        args = self.argument_parser.parse_args(['dem', '-i', '2'])
        self.assertEqual(2, args.increment)

    def test_sv_global_argument(self):
        args = self.argument_parser.parse_args(['dem'])
        self.assertEqual(500, args.sv_global)

        args = self.argument_parser.parse_args(['dem', '--sv_global', '10'])
        self.assertEqual(10, args.sv_global)

        args = self.argument_parser.parse_args(['dem', '-dmax', '20'])
        self.assertEqual(20, args.sv_global)

    def test_sv_local_argument(self):
        args = self.argument_parser.parse_args(['dem'])
        self.assertEqual(100, args.sv_local)

        args = self.argument_parser.parse_args(['dem', '--sv_local', '10'])
        self.assertEqual(10, args.sv_local)

        args = self.argument_parser.parse_args(['dem', '-l', '20'])
        self.assertEqual(20, args.sv_local)

    def test_height_argument(self):
        args = self.argument_parser.parse_args(['dem'])
        self.assertEqual(3, args.height)

        args = self.argument_parser.parse_args(['dem', '--height', '5'])
        self.assertEqual(5, args.height)

        args = self.argument_parser.parse_args(['dem', '-H', '10'])
        self.assertEqual(10, args.height)

    def test_window_argument(self):
        args = self.argument_parser.parse_args(['dem'])
        self.assertEqual(100, args.window)

        args = self.argument_parser.parse_args(['dem', '--window', '30'])
        self.assertEqual(30, args.window)

        args = self.argument_parser.parse_args(['dem', '-W', '50'])
        self.assertEqual(50, args.window)

    def test_var_name_argument(self):
        args = self.argument_parser.parse_args(['dem'])
        self.assertEqual('dem', args.var_name)

        args = self.argument_parser.parse_args(['dem', '--var_name', 'name'])
        self.assertEqual('name', args.var_name)

        args = self.argument_parser.parse_args(['dem', '-N', 'name2'])
        self.assertEqual('name2', args.var_name)

    def test_make_tbreak_argument(self):
        args = self.argument_parser.parse_args(['dem'])
        self.assertEqual(False, args.make_tbreak)

        args = self.argument_parser.parse_args(
            ['dem', '--make_tbreak', 'true'])
        self.assertEqual(True, args.make_tbreak)

        args = self.argument_parser.parse_args(['dem', '-tb', '1'])
        self.assertEqual(True, args.make_tbreak)

    def test_out_tbreak_argument(self):
        args = self.argument_parser.parse_args(['dem'])
        self.assertEqual('./tbreak.nc', args.out_tbreak)

        args = self.argument_parser.parse_args(['dem', '--out_tbreak', 'file'])
        self.assertEqual('file', args.out_tbreak)

        args = self.argument_parser.parse_args(['dem', '-OTB', 'file2'])
        self.assertEqual('file2', args.out_tbreak)


if __name__ == '__main__':
    unittest.main()



import numpy as np

from smrf.envphys.constants import BOIL, FREEZE, SEA_LEVEL


def satw(tk):
    """
    Saturation vapor pressure of water. from IPW satw

    Args:
        tk: temperature in Kelvin

    Returns:
        saturated vapor pressure over water

    20151027 Scott Havens
    """

    # remove bad values
    tk[tk < 0] = np.nan

    l10 = np.log(10.0)

    btk = BOIL/tk
    x = -7.90298*(btk - 1.0) + 5.02808*np.log(btk)/l10 - \
        1.3816e-7*(np.power(10.0, 1.1344e1*(1.0 - tk/BOIL))-1.) + \
        8.1328e-3*(np.power(10.0, -3.49149*(btk - 1.0)) - 1.0) + \
        np.log(SEA_LEVEL)/l10

    x = np.power(10.0, x)

    return x


def sati(tk):
    """
    saturation vapor pressure over ice. From IPW sati

    Args:
        tk: temperature in Kelvin

    Returns:
        saturated vapor pressure over ice

    20151027 Scott Havens
    """

    # remove bad values
    tk[tk < 0] = np.nan

    # preallocate
    x = np.empty(tk.shape)

    # vapor above freezing
    ind = tk > FREEZE
    x[ind] = satw(tk[ind])

    # vapor below freezing
    l10 = np.log(10.0)
    x[~ind] = 100.0 * np.power(10.0, -9.09718*((FREEZE/tk[~ind]) - 1.0) -
                               3.56654*np.log(FREEZE/tk[~ind])/l10 +
                               8.76793e-1*(1.0 - (tk[~ind]/FREEZE)) +
                               np.log(6.1071)/l10)

    return x


def idewpt(vp):
    """
    Calculate the dew point given the vapor pressure

    Args:
        vp - array of vapor pressure values in [Pa]

    Returns:
        dewpt - array same size as vp of the calculated
            dew point temperature [C] (see Dingman 2002).

    """

    # ensure that vp is a numpy array
    vp = np.array(vp)

    # take the log and convert to kPa
    vp = np.log(vp/float(1000))

    # calculate the vapor pressure
    Td = (vp + 0.4926) / (0.0708 - 0.00421*vp)

    return Td


def rh2vp(ta, rh):
    """
    Calculate the vapor pressure given the air temperature
    and relative humidity

    Args:
        ta: array of air temperature in [C]
        rh: array of relative humidity from 0-100 [%]

    Returns:
        vapor pressure
    """

    if rh.flat[0] >= 1.0:
        rh = rh/100.0

    satvp = sati(ta + 273.15)

    return satvp * rh


def satvp(dpt):
    """
    Calculate the saturation vapor pressure at the dew point
    temperature.

    Args:
        dwpt: array of dew point temperature in [C]

    Returns
       vapor_pressure
    """

    return sati(dpt + 273.15)



import copy
import os
import random
import sys
from datetime import datetime
from shutil import copyfile

import numpy as np
import pandas as pd
import pytz
import utm
from inicheck.checkers import CheckType
from inicheck.output import generate_config
from inicheck.utilities import mk_lst
from scipy.interpolate.interpnd import (CloughTocher2DInterpolator,
                                        LinearNDInterpolator)
from scipy.spatial import qhull as qhull

from smrf import __core_config__


class CheckStation(CheckType):
    """
    Custom check for ensuring our stations are always capitalized
    """

    def __init__(self, **kwargs):
        super(CheckStation, self).__init__(**kwargs)

    def type_func(self, value):
        """
        Attempt to convert all the values to upper case.

        Args:
            value: A single string in config entry representing a station name
        Returns:
            value: A single station name all upper case
        """

        return value.upper()


class CheckRawString(CheckType):
    """
    Custom `inicheck` checker that will not change the input string
    """

    def __init__(self, **kwargs):
        super(CheckRawString, self).__init__(**kwargs)

    def type_func(self, value):
        """
        Do not change the passed value at all
        Args:
            value: A single string
        Returns:
            value: A single string unchanged
        """

        return value


def find_configs(directory):
    """
    Searches through a directory and returns all the .ini fulll filenames.

    Args:
        directory: string path to directory.
    Returns:
        configs: list of paths pointing to the config file.
    """

    configs = []
    directory = os.path.abspath(os.path.expanduser(directory))

    for f in os.listdir(directory):
        if f.split('.')[-1] == 'ini':
            configs.append(os.path.join(directory, f))
    return configs


def handle_run_script_options(config_option):
    """
    Handle function for dealing with args in the SMRF run script

    Args:
        config_option: string path to a directory or a specific config file.
    Returns:
        configFile:Full path to an existing config file.
    """
    config_option = os.path.abspath(os.path.expanduser(config_option))

    # User passes a directory
    if os.path.isdir(config_option):
        configs = find_configs(config_option)

        if len(configs) > 1:
            print("\nError: Multiple config files detected in {0} please"
                  " ensure only one is in the folder.\n".format(config_option))
            sys.exit()

        else:
            configFile = configs[0]
    else:
        configFile = config_option

    if not os.path.isfile(configFile):
        print('\nError: Please provide a config file or a directory containing'
              ' one.\n')
        sys.exit()

    return configFile


def nan_helper(y):
    """
    Helper to handle indices and logical indices of NaNs.

    Example:
        >>> # linear interpolation of NaNs
        >>> nans, x= nan_helper(y)
        >>> y[nans]= np.interp(x(nans), x(~nans), y[~nans])

    Args:
        y: 1d numpy array with possible NaNs

    Returns:
        tuple:
            **nans** - logical indices of NaNs
            **index** - a function

    """

    return np.isnan(y), lambda z: z.nonzero()[0]


def set_min_max(data, min_val, max_val):
    """
    Ensure that the data is in the bounds of min and max

    Args:
        data: numpy array of data to be min/maxed
        min_val: minimum threshold to trim data
        max_val: Maximum threshold to trim data

    Returns:
        data: numpy array of data trimmed at min_val and max_val
    """
    if max_val is None:
        max_val = np.inf
    if min_val is None:
        min_val = -np.inf

    ind = np.isnan(data)

    data[data <= min_val] = min_val
    data[data >= max_val] = max_val

    data[ind] = np.nan

    return data


def water_day(indate):
    """
    Determine the decimal day in the water year

    Args:
        indate: datetime object

    Returns:
        tuple:
            **dd** - decimal day from start of water year
            **wy** - Water year

    20160105 Scott Havens
    """
    tp = indate.timetuple()

    # create a test start of the water year
    test_date = datetime(tp.tm_year, 10, 1, 0, 0, 0)
    test_date = test_date.replace(tzinfo=pytz.timezone(indate.tzname()))

    # check to see if it makes sense
    if indate < test_date:
        wy = tp.tm_year
    else:
        wy = tp.tm_year + 1

    # actual water year start
    wy_start = datetime(wy-1, 10, 1, 0, 0, 0)
    wy_start = wy_start.replace(tzinfo=pytz.timezone(indate.tzname()))

    # determine the decimal difference
    d = indate - wy_start
    dd = d.days + d.seconds/86400.0

    return dd, wy


def is_leap_year(year):
    return (year % 4 == 0 and year % 100 != 0) or year % 400 == 0


def backup_input(data, config_obj):
    """
    Backs up input data files so a user can rerun a run with the exact data
    used for a run.

    Args:
        data: Pandas dataframe containing the station data
        config_obj: The config object produced by inicheck
    """
    # mask copy
    backup_config_obj = copy.deepcopy(config_obj)

    # Make the output dir
    backup_dir = os.path.join(backup_config_obj.cfg['output']['out_location'],
                              'input_backup')
    if not os.path.isdir(backup_dir):
        os.mkdir(backup_dir)

    # Check config file for csv section and remove alternate data form config
    if 'csv' not in backup_config_obj.cfg.keys():
        backup_config_obj.cfg['csv'] = {}
        # With a new section added, we need to remove the other data sections
        # backup_config_obj.apply_recipes()

    if 'stations' in backup_config_obj.cfg.keys():
        if 'client' in backup_config_obj.cfg['stations']:
            del backup_config_obj.cfg['stations']['client']

    # Output station data to CSV
    csv_var = ['metadata', 'air_temp', 'vapor_pressure', 'precip',
               'wind_speed', 'wind_direction', 'cloud_factor']

    for k in csv_var:
        fname = os.path.join(backup_dir, k + '.csv')
        v = getattr(data, k)
        v.to_csv(fname)

        # Adjust and output the inifile
        backup_config_obj.cfg['csv'][k] = fname

    # Copy topo files over to backup
    ignore = ['northern_hemisphere',
              'gradient_method', 'sky_view_factor_angles']
    for s in backup_config_obj.cfg['topo'].keys():
        src = backup_config_obj.cfg['topo'][s]
        # make not a list if lenth is 1
        if isinstance(src, list):
            src = mk_lst(src, unlst=True)
        # Avoid attempring to copy files that don't exist
        if s not in ignore and src is not None:
            dst = os.path.join(backup_dir, os.path.basename(src))
            backup_config_obj.cfg["topo"][s] = dst
            copyfile(src, dst)

    # We dont want to backup the backup
    backup_config_obj.cfg['output']['input_backup'] = False

    # Output inifile
    generate_config(backup_config_obj, os.path.join(
        backup_dir, 'backup_config.ini'))


def check_station_colocation(metadata_csv=None, metadata=None):
    """
    Takes in a data frame representing the metadata for the weather stations
    as produced by :mod:`smrf.framework.model_framework.SMRF.loadData` and
    check to see if any stations have the same location.

    Args:
        metadata_csv: CSV containing the metdata for weather stations
        metadata: Pandas Dataframe containing the metdata for weather stations

    Returns:
        repeat_sta: list of station primary_id that are colocated
    """
    if metadata_csv is not None:
        metadata = pd.read_csv(metadata_csv)
        metadata.set_index('primary_id', inplace=True)

    # Unique station locations
    unique_x = list(metadata.xi.unique())
    unique_y = list(metadata.yi.unique())

    repeat_sta = []

    # Cycle through all the positions look for multiple  stations at a position
    for x in unique_x:
        for y in unique_y:
            x_search = metadata['xi'] == x
            y_search = metadata['yi'] == y
            stations = metadata.index[x_search & y_search].tolist()

            if len(stations) > 1:
                repeat_sta.append(stations)

    if len(repeat_sta) == 0:
        repeat_sta = None

    return repeat_sta


def get_config_doc_section_hdr():
    """
    Returns the header dictionary for linking modules in smrf to the
    documentation generated by inicheck auto doc functions
    """
    hdr_dict = {}

    dist_modules = ['air_temp', 'vapor_pressure', 'precip', 'wind', 'albedo',
                    'thermal', 'solar', 'cloud_factor', 'soil_temp']

    for d in dist_modules:
        if d == 'precip':
            sec = 'precipitation'
        else:
            sec = d

        # If distributed module link api
        intro = ("The {0} section controls all the available parameters that"
                 " effect the distribution of the {0} module, espcially  the"
                 " associated models. For more detailed information please see"
                 " :mod:`smrf.distribute.{0}`\n").format(sec)

        hdr_dict[d] = intro

    return hdr_dict


def get_asc_stats(fp):
    """
    Returns header of ascii dem file
    """
    ts = {}
    header = {}

    ff = open(fp, 'r')
    for idl, line in enumerate(ff):
        tmp_line = line.strip().split()
        header[tmp_line[0]] = tmp_line[1]
        if idl >= 5:
            break
    ff.close()

    ts['nx'] = int(header['ncols'])
    ts['ny'] = int(header['nrows'])
    ts['du'] = float(header['cellsize'])
    ts['dv'] = float(header['cellsize'])
    ts['u'] = float(header['yllcorner'])
    ts['v'] = float(header['xllcorner'])

    ts['x'] = ts['v'] + ts['dv']*np.arange(ts['nx'])
    ts['y'] = ts['u'] + ts['du']*np.arange(ts['ny'])
    ts['y'] = ts['y'][::-1]

    # ASCII are lower left coordiante
    # must shift to cell center for interpolation
    ts['x'] = ts['x'] + ts['dv']/2
    ts['y'] = ts['y'] + ts['du']/2

    return ts


def getqotw():
    p = os.path.dirname(__core_config__)
    q_f = os.path.abspath(os.path.join('{0}'.format(p), '.qotw'))
    with open(q_f) as f:
        qs = f.readlines()
        f.close()
    i = random.randrange(0, len(qs))
    return qs[i]


def interp_weights(xy, uv, d=2):
    """
    Find vertices and weights of LINEAR interpolation for gridded interp.
    This routine follows the methods of scipy.interpolate.griddata as outlined
    here:
    https://stackoverflow.com/questions/20915502/speedup-scipy-griddata-for-multiple-interpolations-between-two-irregular-grids
    This function finds the vertices and weights which is the most
    computationally expensive part of the routine. The interpolateion can
    then be done quickly.

    Args:
        xy: n by 2 array of flattened meshgrid x and y coords of WindNinja grid
        uv: n by 2 array of flattened meshgrid x and y coords of SMRF grid
        d:  dimensions of array (i.e. 2 for our purposes)

    Returns:
        vertices:
        wts:

    """
    tri = qhull.Delaunay(xy)
    simplex = tri.find_simplex(uv)
    vertices = np.take(tri.simplices, simplex, axis=0)
    temp = np.take(tri.transform, simplex, axis=0)
    delta = uv - temp[:, d]
    bary = np.einsum('njk,nk->nj', temp[:, :d, :], delta)

    return vertices, np.hstack((bary, 1 - bary.sum(axis=1, keepdims=True)))


def grid_interpolate(values, vtx, wts, shp, fill_value=np.nan):
    """
    Broken out gridded interpolation from scipy.interpolate.griddata that takes
    the vertices and wts from interp_weights function

    Args:
        values: flattened WindNinja wind speeds
        vtx:    vertices for interpolation
        wts:    weights for interpolation
        shape:  shape of SMRF grid
        fill_value: value for extrapolated points

    Returns:
        ret:    interpolated values
    """
    ret = np.einsum('nj,nj->n', np.take(values, vtx), wts)
    ret[np.any(wts < 0, axis=1)] = fill_value

    ret = ret.reshape(shp[0], shp[1])

    return ret


def grid_interpolate_deconstructed(tri, values, grid_points, method='linear'):
    """
    Underlying methods from scipy grid_data broken out to pass in the tri
    values returned from qhull.Delaunay. This is done to improve the speed
    of using grid_data

    Args:
        tri:            values returned from qhull.Delaunay
        values:         values at HRRR stations generally
        grid_points:    tuple of vectors for X,Y coords of grid stations
        method:         either linear or cubic

    Returns:
        result of interpolation to gridded points
    """

    if method == 'cubic':
        return CloughTocher2DInterpolator(tri, values)(grid_points)
    elif method == 'linear':
        return LinearNDInterpolator(tri, values)(grid_points)


def apply_utm(s, force_zone_number):
    """
    Calculate the utm from lat/lon for a series

    Args:
        s: pandas series with fields latitude and longitude
        force_zone_number: default None, zone number to force to

    Returns:
        s: pandas series with fields 'X' and 'Y' filled
    """
    p = utm.from_latlon(s.latitude, s.longitude,
                        force_zone_number=force_zone_number)
    s['utm_x'] = p[0]
    s['utm_y'] = p[1]
    return s


def date_range(start_date, end_date, increment, timezone):
    """
    Calculate a list between start and end date with
    an increment
    """

    return list(pd.date_range(
        start_date,
        end_date,
        freq="{}min".format(increment),
        tz=timezone
    ))



import pytz
from pandas import to_datetime

from smrf.framework.model_framework import SMRF
from smrf.tests.smrf_test_case import SMRFTestCase


class TestModelFramework(SMRFTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.smrf = SMRF(cls.config_file)

    def test_start_date(self):
        self.assertEqual(
            self.smrf.start_date,
            to_datetime(self.smrf.config['time']['start_date'], utc=True)
        )

    def test_end_date(self):
        self.assertEqual(
            self.smrf.end_date,
            to_datetime(self.smrf.config['time']['end_date'], utc=True)
        )

    def test_time_zone(self):
        self.assertEqual(self.smrf.time_zone, pytz.UTC)

    def test_date_time(self):
        self.assertEqual(
            self.smrf.date_time[0],
            to_datetime('1998-01-14 15:00:00', utc=True)
        )
        self.assertEqual(
            self.smrf.date_time[-1],
            to_datetime('1998-01-14 19:00:00', utc=True)
        )
        self.assertEqual(
            self.smrf.date_time[0].tzname(),
            str(pytz.UTC)
        )
        self.assertEqual(
            type(self.smrf.date_time),
            list
        )

    def test_assert_time_steps(self):
        self.assertEqual(self.smrf.time_steps, 5)


class TestModelFrameworkMST(SMRFTestCase):
    """
    Test timezone handling for MST.
    """
    TIMEZONE = pytz.timezone('MST')

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        base_config = cls.base_config_copy()
        base_config.cfg['time']['time_zone'] = 'MST'
        cls.smrf = SMRF(base_config)

    def test_timezone_error(self):
        base_config = self.base_config
        base_config.cfg['time']['time_zone'] = 'mst'
        with self.assertRaises(Exception):
            SMRF(base_config)

    def test_start_date(self):
        self.assertEqual(
            self.smrf.start_date,
            to_datetime(self.smrf.config['time']['start_date']).tz_localize(
                self.TIMEZONE
            )
        )

    def test_end_date(self):
        self.assertEqual(
            self.smrf.end_date,
            to_datetime(self.smrf.config['time']['end_date']).tz_localize(
                self.TIMEZONE
            )
        )

    def test_time_zone(self):
        self.assertEqual(self.smrf.time_zone, self.TIMEZONE)

    def test_date_time(self):
        self.assertEqual(
            self.smrf.date_time[0],
            to_datetime('1998-01-14 15:00:00').tz_localize(self.TIMEZONE)
        )
        self.assertEqual(
            self.smrf.date_time[-1],
            to_datetime('1998-01-14 19:00:00').tz_localize(self.TIMEZONE)
        )
        self.assertEqual(
            self.smrf.date_time[0].tz.zone,
            self.TIMEZONE.zone
        )



import numpy as np


def twostream(cosz, S0, tau=0.2, omega=0.85, g=0.3, R0=0.5):
    """
    Provides twostream solution for single-layer atmosphere over horizontal
    surface, using solution method in: Two-stream approximations to radiative
    transfer in planetary atmospheres: a unified description of existing
    methods and a new improvement, Meador & Weaver, 1980, or will use the
    delta-Eddington  method, if the -d flag is set (see: Wiscombe & Joseph
    1977).

    Args:
        cosz: The cosine of the incidence angle is cos (from program sunang).
            An error if cosz is <= 0.0; set all outputs to 0.0 and
            go on. Program will fail if incidence angle is <= 0.0, unless -0
            has been set.
        S0: The direct beam irradiance is S0 This is usually the solar
            constant for the specified wavelength band, on the specified date,
            at the top of the atmosphere, from radiation.solar.
        tau: The optical depth is tau.  0 implies an infinite optical depth.
        omega: The single-scattering albedo
        g: The asymmetry factor is g.
        R0: The reflectance of the substrate is R0.  If R0 is negative, it
            will be set to zero.

    Returns:
        dictionary with the following keys
        R[0] - reflectance
        R[1] - transmittance
        R[2] - direct transmittance
        R[3] - upwelling irradiance
        R[4] - total irradiance at bottom
        R[5] - direct irradiance normal to beam
    """

    if cosz <= 0:
        return {
            'reflectance': 0,
            'transmittance': 0,
            'direct_transmittance': 0,
            'upwelling_irradiance': 0,
            'irradiance_at_bottom': 0,
            'irradiance_normal_to_beam': 0
        }

    if S0 <= 0:
        raise ValueError('The direct beam irradiance (S0) is less than 0')

    if isinstance(tau, float):
        tau = np.array([tau])

    idx = tau == 0
    tau[idx] = 1e15

    if R0 < 0:
        R0 = 0

    # gamma's for phase function for all input
    gamma = mwgamma(cosz, omega, g)

    # Calcuate twostream (from libmodel/radiation/twostream.c)
    gam1 = gamma[0]
    gam2 = gamma[1]
    gam3 = gamma[2]
    gam4 = gamma[3]
    alph1 = gam1 * gam4 + gam2 * gam3
    alph2 = gam2 * gam4 + gam1 * gam3
    if gam1 < gam2:
        raise ValueError("gam1 ({}) < gam2 ({})".format(gam1, gam2))

    # (Hack - gam1 = gam2 with conservative scattering.  Need to fix this.
    # Ruh-roh...
    if gam1 == gam2:
        gam1 += 2.2204460492503131e-16

    xi = np.sqrt((gam1 - gam2) * (gam2 + gam1))
    em = np.exp(-tau * xi)
    et = np.exp(-tau / cosz)
    ep = np.exp(tau * xi)
    gpx = xi + gam1
    opx = cosz * xi + 1

    # more intermediate variables, needed only for finite case
    omx = 1 - cosz * xi
    gmx = gam1 - xi
    rm = gam2 - gmx * R0
    rp = gam2 - gpx * R0

    # denominator for reflectance and transmittance
    denrt = ep * gpx * rm - em * gmx * rp

    # reflectance
    refl = (omega * (ep * rm * (gam3 * xi + alph2) / opx
                     - em * rp * (alph2 - gam3 * xi) / omx)
            + 2 * et * gam2
            * (R0 - ((alph1 * R0 - alph2) * cosz + gam4
                     * R0 + gam3) * omega / (omx * opx)) * xi) / denrt

    # transmittance
    trans = (et * (ep * gpx * (gam2 - omega * (alph2
                                               - gam3 * xi) / omx)
                   - em * gmx * (gam2 - omega * (gam3 * xi + alph2) / opx))
             + 2 * gam2 * (alph1 * cosz + gam4) * omega * xi /
             (omx * opx)) / denrt

    # direct transmittance
    btrans = et

    # semi-infinite?
    idx = np.argwhere(((em == 0) & (et == 0)) | (ep >= 1e15))
    refl[idx] = omega * (gam3 * xi + alph2) / (gpx * opx)
    btrans[idx] = 0
    trans[idx] = 0

    assert(np.min(refl) >= 0)
    assert(np.min(trans) >= 0)
    assert(np.min(btrans) >= 0)
    assert(np.all(trans >= btrans * cosz))

    return {
        'reflectance': refl,
        'transmittance': trans,
        'direct_transmittance': btrans,
        'upwelling_irradiance': refl * cosz * S0,
        'irradiance_at_bottom': trans * cosz * S0,
        'irradiance_normal_to_beam': btrans * S0
    }


def mwgamma(cosz, omega, g):
    """
    gamma's for phase function for input using the MEADOR WEAVER method

    Two-stream approximations to radiative transfer in planetary
    atmospheres: a unified description of existing methods and a
    new improvement, Meador & Weaver, 1980

    Args:
        cosz: cosine illumination angle
        omega: single-scattering albedo
        g: scattering asymmetry param

    Returns
        gamma values
    """

    # Meador-Weaver hybrid approx
    b0 = beta_0(cosz, g)

    # hd is denom for hybrid g1,g2
    hd = 4 * (1 - g * g * (1 - cosz))

    # these are Horner expressions corresponding to Meador-Weaver Table 1)
    if g == 1:
        g1 = g2 = (g * ((3 * (g - 1) + 4 * b0) * g - 3) + 3) / hd
    else:
        g1 = (g * (g * ((3 * g + 4 * b0) * omega - 3) -
                   3 * omega) - 4 * omega + 7) / hd

        g2 = (g * (g * ((3 * g + 4 * (b0 - 1)) * omega + 1) -
                   3 * omega) + 4 * omega - 1) / hd

    g3 = b0
    g4 = 1 - g3
    g = [g1, g2, g3, g4]
    return g


def beta_0(cosz, g):
    """
    we find the integral-sum

    sum (n=0 to inf) g^n * (2*n+1) * Pn(u0) * int (u'=0 to 1) Pn(u')

    note that int of Pn vanishes for even values of n (Abramowitz &
    Stegun, eq 22.13.8-9); therefore the series becomes

    sum (n=0 to inf) g^n * (2*n+1) * Pn(u0) * f(m)

    where 2*m+1 = n and the f's are computed recursively

    Args:
        cosz: cosine illumination angle
        g: scattering asymmetry param

    Returns:
        beta_0

    """

    MAXNO = 2048
    TOL = 1e-9

    assert(cosz > 0 and cosz <= 1)
    assert(g >= 0 and g <= 1)

    # Legendre polynomials of degree 0 and 1

    pnm2 = 1
    pnm1 = cosz

    # first coefficients and initial sum
    fm = -1/8
    gn = 7 * g * g * g
    the_sum = 3 * g * cosz / 2

    # sum until convergence; we use the even terms only for the recursive
    # calculation of the Legendre polynomials
    if (g != 0 and cosz != 0):
        for n in range(2, MAXNO):

            # order n Legendre polynomial
            pn = ((2 * n - 1) * cosz * pnm1 + (1 - n) * pnm2) / n

            # even terms vanish
            if (n % 2) == 1:
                last = the_sum
                the_sum = last + gn * fm * pn
                if (np.abs((the_sum - last) / the_sum) < TOL and the_sum <= 1):
                    break

                # recursively find next f(m) and gn coefficients
                # n = 2 * m + 1
                m = (n - 1) / 2
                fm *= -(2 * m + 1) / (2 * (m + 2))
                gn *= g * g * (4 * m + 7) / (4 * m + 3)

            # ready to compute next Legendre polynomial
            pnm2 = pnm1
            pnm1 = pn

        # warn if no convergence
        if n == MAXNO:
            print("%s: %s - cosz=%g g=%g sum=%g last=%g fm=%g",
                  "betanaught", "no convergence",
                  cosz, g, the_sum, last, fm)
            if the_sum > 1:
                the_sum = 1

        else:
            assert(the_sum >= 0 and the_sum <= 1)

    else:
        the_sum = 0

    return (1 - the_sum) / 2



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os

from setuptools import Extension, find_packages, setup
from setuptools.command.build_ext import build_ext as _build_ext

# Test if compiling with cython or using the C source
try:
    from Cython.Distutils import build_ext as _build_ext
except ImportError:
    USE_CYTHON = False
else:
    USE_CYTHON = True

print('Using Cython {}'.format(USE_CYTHON))
ext = '.pyx' if USE_CYTHON else '.c'


def c_name_from_path(location, name):
    return os.path.join(location, name).replace('/', '.')


class build_ext(_build_ext):
    def finalize_options(self):
        _build_ext.finalize_options(self)
        # Prevent numpy from thinking it is still in its setup process:
        __builtins__.__NUMPY_SETUP__ = False
        import numpy
        self.include_dirs.append(numpy.get_include())


# Give user option to specify his local compiler name
if "CC" not in os.environ:
    os.environ["CC"] = "gcc"

ext_modules = []
extension_params = dict(
    extra_compile_args=['-fopenmp', '-O3'],
    extra_link_args=['-fopenmp', '-O3'],
)

# detrended kriging
source_folder = 'smrf/spatial/dk'
ext_modules += [
    Extension(
        c_name_from_path(source_folder, 'detrended_kriging'),
        sources=[os.path.join(source_folder, val) for val in [
            "detrended_kriging" + ext,
            "krige.c",
            "lusolv.c",
            "array.c"
        ]],
        **extension_params
    ),
]

# envphys core c functions
source_folder = 'smrf/envphys/core'
ext_modules += [
    Extension(
        c_name_from_path(source_folder, 'envphys_c'),
        sources=[os.path.join(source_folder, val) for val in [
            "envphys_c" + ext,
            "topotherm.c",
            "dewpt.c",
            "iwbt.c"
        ]],
        **extension_params
    ),
]

# wind model c functions
source_folder = 'smrf/utils/wind'
ext_modules += [
    Extension(
        c_name_from_path(source_folder, 'wind_c'),
        sources=[os.path.join(source_folder, val) for val in [
            "wind_c" + ext,
            "breshen.c",
            "calc_wind.c"
        ]],
        **extension_params
    ),
]

cmdclass = {}
if USE_CYTHON:
    cmdclass = {'build_ext': build_ext}

with open('requirements.txt') as requirements_file:
    requirements = requirements_file.read()

with open('README.md') as readme_file:
    readme = readme_file.read()

setup(
    name='smrf-dev',
    description="Distributed snow modeling for water resources",
    author="Scott Havens",
    author_email='scott.havens@ars.usda.gov',
    url='https://github.com/USDA-ARS-NWRC/smrf',
    long_description=readme,
    long_description_content_type="text/markdown",
    packages=find_packages(include=['smrf', 'smrf.*']),
    install_requires=requirements,
    python_requires='>3.5',
    include_package_data=True,
    package_data={
        'smrf': [
            './framework/CoreConfig.ini',
            './framework/.qotw',
            './framework/recipes.ini',
            './framework/changelog.ini'
        ]
    },
    license="CC0 1.0",
    zip_safe=False,
    keywords='smrf',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ],
    test_suite='smrf.tests',
    # tests_require=test_requirements,
    cmdclass=cmdclass,
    ext_modules=ext_modules,
    scripts=[
        'scripts/update_configs',
        'scripts/run_smrf',
        'scripts/gen_maxus'
    ],
    extras_require={
        'docs': [
            'Sphinx>=3.0,<=4.0',
            'pydata-sphinx-theme',
            'sphinxcontrib-bibtex>=1.0',
            'sphinxcontrib-websupport>=1.0.1',
        ],
        'tests': [
            'mock',
        ],
    },
    use_scm_version={
        'local_scheme': 'node-and-date',
    },
    setup_requires=[
        'setuptools_scm',
        'numpy'
    ],
)



# -*- coding: utf-8 -*-
# flake8: noqa
from .csv import InputCSV
from .hrrr_grib import InputGribHRRR
from .load_topo import Topo
from .netcdf import InputNetcdf
from .wrf import InputWRF

from .load_data import InputData  # isort:skip



import numpy as np


# Coefficients for snow relationship
class Coefficients:
    TR0 = 0.5
    PCR0 = 0.25
    PC0 = 0.75


def calc_percent_snow(tpp, t_max=0.0):
    """
    Calculates the percent snow for the nasde_models piecewise_susong1999
    and marks2017.

    Args:
        tpp: A numpy array of temperature, use dew point temperature if
            available [degree C].
        t_max: Max temperature that the percent snow is estimated.
              Default is 0.0 Degrees C.

    Returns:
        numpy.array:
        A fraction of the precipitation at each pixel that is snow provided
        by tpp.
    """

    percent = np.zeros(tpp.shape)

    percent[tpp <= -0.5] = 1.0

    ind = (tpp > -0.5) & (tpp <= 0.0)
    if np.any(ind):
        percent[ind] = (-tpp[ind] / Coefficients.TR0) * \
                       Coefficients.PCR0 + Coefficients.PC0

    ind = (tpp > 0.0) & (tpp <= t_max + 1.0)
    if np.any(ind):
        percent[ind] = (-tpp[ind] / (t_max + 1.0)) * \
                       Coefficients.PC0 + Coefficients.PC0

    return percent


def check_temperature(tpp, t_max=0.0, t_min=-10.0):
    """
    Sets  the precipitation temperature and snow temperature.

    Args:
        tpp: A numpy array of temperature, use dew point temperature
            if available [degrees C].
        t_max: Thresholds the max temperature of the snow [degrees C].
        t_min: Minimum temperature that the precipitation temperature
            [degrees C].

    Returns:
        tuple:
           - **tpp** (*numpy.array*) - Modified precipitation temperature
                     that limited to a minimum set by t_min.
           - **t_snow** (*numpy.array*) - Temperature of the surface of the
                         snow set by the precipitation temperature and
                         limited by t_max, where t_snow > t_max = t_max.

    """

    tpp[tpp < t_min] = t_min

    t_snow = tpp.copy()
    t_snow[tpp > t_max] = t_max

    return tpp, t_snow



from copy import deepcopy

from inicheck.tools import cast_all_variables

from smrf.framework.model_framework import run_smrf
from smrf.tests.smrf_test_case import SMRFTestCase


class TestLoadGrid(SMRFTestCase):

    def test_grid_wrf(self):
        """ WRF NetCDF loading """

        config = deepcopy(self.base_config)
        del config.raw_cfg['csv']

        adj_config = {
            'gridded': {
                'data_type': 'wrf',
                'wrf_file': './gridded/WRF_test.nc',
            },
            'system': {
                'threading': 'False',
                'log_file': './output/log.txt'
            },
            'precip': {
                'station_adjust_for_undercatch': 'False'
            },
            'wind': {
                'wind_model': 'interp'
            },
            'thermal': {
                'correct_cloud': 'False',
                'correct_veg': 'True'
            },
            'time': {
                'start_date': '2015-03-03 00:00',
                'end_date': '2015-03-03 04:00'
            }
        }

        config.raw_cfg.update(adj_config)

        # set the distribution to grid, thermal defaults will be fine
        for v in self.dist_variables:
            config.raw_cfg[v]['grid_mask'] = 'False'

        # fix the time to that of the WRF_test.nc
        config.apply_recipes()
        config = cast_all_variables(config, config.mcfg)

        # ensure that the recipes are used
        self.assertTrue(
            'station_adjust_for_undercatch' not in config.cfg['precip'].keys()
        )
        self.assertFalse(config.cfg['thermal']['correct_cloud'])
        self.assertTrue(config.cfg['thermal']['correct_veg'])

        run_smrf(config)



import numpy as np
import pandas as pd
from pykrige.ok import OrdinaryKriging


class KRIGE:
    '''
    Kriging class based on the pykrige package
    '''

    def __init__(self, mx, my, mz, GridX, GridY, GridZ, config):
        """
        Args:
            mx: x locations for the points
            my: y locations for the points
            GridX: x locations in grid to interpolate over
            GridY: y locations in grid to interpolate over
            power: power of the inverse distance weighting
        """

        # Measurement point locations
        self.mx = mx
        self.my = my
        self.mz = mz
        self.nsta = len(mx)

        # grid information
        self.GridX = GridX
        self.GridY = GridY
        self.GridZ = GridZ
        self.ngrid = GridX.shape[0] * GridX.shape[1]

        # data information
        self.data = None
        self.nan_val = []

        self.config = config

        # kriging parameters for pykrige
        self.variogram_model = self.config['krig_variogram_model']  # 'linear'
        self.variogram_parameters = None
        self.variogram_function = None
        # np.min([np.round(len(mx)/2), 6])
        self.nlags = self.config['krig_nlags']
        self.weight = self.config['krig_weight']
        self.anisotropy_scaling = self.config['krig_anisotropy_scaling']  # 1.0
        self.anisotropy_angle = self.config['krig_anisotropy_angle']  # 0.0
        self.verbose = False
        self.enable_plotting = False
        self.enable_statistics = False
        # not in the pypi release of PyKrige
        self.coordinates_type = self.config['krig_coordinates_type']

        # pykrige execution
        self.backend = 'vectorized'
        self.n_closest_points = None

    def calculate(self, data):
        """
        Estimate the variogram, calculate the model, then apply to the grid

        Arg:
            data: numpy array same length as m*
            config: configuration for dk

        Returns:
            v: Z-values of specified grid or at thespecified set of points.
               If style was specified as 'masked', zvalues will
               be a numpy masked array.
            sigmasq: Variance at specified grid points or
                     at the specified set of points. If style was specified as
                     'masked', sigmasq will be a numpy masked array.
        """

        nan_val = pd.isnull(data)

        if self.config['detrend']:
            d = self.detrendData(data)
        else:
            d = data.copy()

        OK = OrdinaryKriging(self.mx[~nan_val],
                             self.my[~nan_val],
                             d[~nan_val],
                             variogram_model=self.variogram_model,
                             variogram_parameters=self.variogram_parameters,
                             variogram_function=self.variogram_function,
                             nlags=self.nlags,
                             weight=self.weight,
                             anisotropy_scaling=self.anisotropy_scaling,
                             anisotropy_angle=self.anisotropy_angle,
                             verbose=self.verbose,
                             enable_plotting=self.enable_plotting,
                             enable_statistics=self.enable_statistics)

        v, ss1 = OK.execute('grid',
                            self.GridX[0, :],
                            self.GridY[:, 0],
                            backend=self.backend,
                            n_closest_points=self.n_closest_points)

        if self.config['detrend']:
            # retrend the residuals
            v = self.retrendData(v)

        return v, ss1

    def detrendData(self, data, flag=0, zeros=None):
        '''
        Detrend the data in val using the heights zmeas

        Args:
            data: is the same size at mx,my
            flag: - 1 for positive, -1 for negative, 0 for any trend imposed

        Returns:
            data minus the elevation trend

        '''

        # calculate the trend on any real data
        nan_val = np.isnan(data)
        pv = np.polyfit(self.mz[~nan_val], data[~nan_val], 1)

        # apply trend constraints
        if flag == 1 and pv[0] < 0:
            pv = np.array([0, 0])
        elif (flag == -1 and pv[0] > 0):
            pv = np.array([0, 0])

        self.pv = pv

        # detrend the data
        el_trend = self.mz * pv[0] + pv[1]

        if zeros is not None:
            data[zeros] = self.zeroVal

        return data - el_trend

    def retrendData(self, idw):
        '''
        Retrend the IDW values
        '''

        # retrend the data
        return idw + self.pv[0]*self.GridZ + self.pv[1]



# cython: language_level=3
"""
Cython wrapper to the underlying C code

20160816
"""
import cython
import numpy as np
cimport numpy as np

# Numpy must be initialized. When using numpy from C or Cython you must
# _always_ do that, or you will have segfaults
np.import_array()


cdef extern from "wind_header.h":
    void calc_maxus(int nx, int ny, double *x, double *y, double *z, double *X_start, double *Y_start, 
                    double *X_end, double *Y_end, double height, int nthreads, double *maxus);


@cython.boundscheck(False)
@cython.wraparound(False)
# https://github.com/cython/cython/wiki/tutorials-NumpyPointerToC
def call_maxus(x, y, Z, X, Y, Xi, Yi, double height=3, int nthreads=1):
    '''
    Call the function maxus_grid in calc_wind.c which will iterate over the grid
    within the C code

    Args:
        ad - [nsta x nsta] matrix of distances between stations
        dgrid - [ngrid x nsta] matrix of distances between grid points and stations
        elevations - [nsta] array of station elevations
        weights (return) - [ngrid x nsta] matrix of kriging weights calculated
        nthreads - number of threads to use in parallel processing

    Out:
        weights changed in place

    20160222 Scott Havens
    '''

    cdef int nx, ny
    ny, nx = X.shape[0], X.shape[1]

    # convert the x array to C
    cdef np.ndarray[double, mode="c", ndim=1] x_arr
    x_arr = np.ascontiguousarray(x, dtype=np.float64)

    # convert the x array to C
    cdef np.ndarray[double, mode="c", ndim=1] y_arr
    y_arr = np.ascontiguousarray(y, dtype=np.float64)

    # convert the z array to C
    cdef np.ndarray[double, mode="c", ndim=2] elev
    elev = np.ascontiguousarray(Z, dtype=np.float64)

    # convert the X array to C
    cdef np.ndarray[double, mode="c", ndim=2] X_start
    X_start = np.ascontiguousarray(X, dtype=np.float64)

    # convert the Y array to C
    cdef np.ndarray[double, mode="c", ndim=2] Y_start
    Y_start = np.ascontiguousarray(Y, dtype=np.float64)

    # convert the Xi array to C
    cdef np.ndarray[double, mode="c", ndim=2] X_end
    X_end = np.ascontiguousarray(Xi, dtype=np.float64)

    # convert the Yi array to C
    cdef np.ndarray[double, mode="c", ndim=2] Y_end
    Y_end = np.ascontiguousarray(Yi, dtype=np.float64)

    # convert the dgrid to C
    z = np.zeros_like(X)
    cdef np.ndarray[double, mode="c", ndim=2] maxus
    maxus = np.ascontiguousarray(z, dtype=np.float64)

    calc_maxus(nx, ny, &x_arr[0], &y_arr[0], &elev[0,0], &X_start[0,0], &Y_start[0,0],
                &X_end[0,0], &Y_end[0,0], height, nthreads, &maxus[0,0])

    return maxus



from copy import deepcopy

from inicheck.tools import cast_all_variables

from smrf.framework.model_framework import run_smrf
from smrf.tests.smrf_test_case import SMRFTestCase


class TestLoadHRRR(SMRFTestCase):

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        config = cls.base_config_copy()
        del config.raw_cfg['csv']

        adj_config = {
            'gridded': {
                'data_type': 'hrrr_grib',
                'hrrr_directory': './gridded/hrrr_test/',
            },
            'time': {
                'start_date': '2018-07-22 16:00',
                'end_date': '2018-07-22 20:00',
                'time_zone': 'utc'
            },
            'system': {
                'threading': False,
                'log_file': './output/test.log'
            },
            'air_temp': {
                'grid_local': True,
                'grid_local_n': 25
            },
            'vapor_pressure': {
                'grid_local': True,
                'grid_local_n': 25
            },
            'precip': {
                'grid_local': True,
                'grid_local_n': 25,
                'precip_temp_method': 'dew_point'
            },
            'wind': {
                'wind_model': 'interp'
            },
            'thermal': {
                'correct_cloud': True,
                'correct_veg': True
            },
            'albedo': {
                'grain_size': 300.0,
                'max_grain': 2000.0
            }
        }
        config.raw_cfg.update(adj_config)

        # set the distribution to grid, thermal defaults will be fine
        for v in cls.DIST_VARIABLES:
            config.raw_cfg[v]['distribution'] = 'grid'
            config.raw_cfg[v]['grid_mask'] = 'False'

        config.apply_recipes()
        config = cast_all_variables(config, config.mcfg)

        cls.config = config
        cls.gold_dir = cls.basin_dir.joinpath('gold_hrrr')

    def test_grid_hrrr_local(self):

        run_smrf(self.config)
        self.compare_hrrr_gold()

    def test_load_timestep(self):

        config = deepcopy(self.config)
        config.raw_cfg['gridded']['hrrr_load_method'] = 'timestep'
        config.apply_recipes()
        config = cast_all_variables(config, config.mcfg)

        run_smrf(config)

        self.compare_hrrr_gold()

    def test_load_timestep_threaded(self):

        config = deepcopy(self.config)
        config.raw_cfg['gridded']['hrrr_load_method'] = 'timestep'
        config.raw_cfg['system']['threading'] = True
        config.raw_cfg['system']['queue_max_values'] = 1
        config.apply_recipes()
        config = cast_all_variables(config, config.mcfg)

        run_smrf(config)

        self.compare_hrrr_gold()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from smrf.framework.model_framework import run_smrf
from smrf.tests.smrf_test_case_lakes import SMRFTestCaseLakes


class TestLakes(SMRFTestCaseLakes):
    """
    Integration test for SMRF without threading.
    Runs the short simulation over Lakes.
    """
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.smrf = run_smrf(cls.config_file)

    def test_air_temp(self):
        self.compare_netcdf_files('air_temp.nc')

    def test_precip_temp(self):
        self.compare_netcdf_files('precip_temp.nc')

    def test_net_solar(self):
        self.compare_netcdf_files('net_solar.nc')

    def test_percent_snow(self):
        self.compare_netcdf_files('percent_snow.nc')

    def test_precip(self):
        self.compare_netcdf_files('precip.nc')

    def test_thermal(self):
        self.compare_netcdf_files('thermal.nc')

    def test_wind_speed(self):
        self.compare_netcdf_files('wind_speed.nc')

    def test_wind_direction(self):
        self.compare_netcdf_files('wind_direction.nc')

    def test_snow_density(self):
        self.compare_netcdf_files('snow_density.nc')

    def test_vapor_pressure(self):
        self.compare_netcdf_files('vapor_pressure.nc')


class TestLakesThreaded(SMRFTestCaseLakes):
    """
    Integration test for SMRF.
    Runs the short simulation over Lakes.
    """
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        cls.smrf = run_smrf(cls.thread_config())

    def tearDown(self):
        pass

    def test_air_temp(self):
        self.compare_netcdf_files('air_temp.nc')

    def test_precip_temp(self):
        self.compare_netcdf_files('precip_temp.nc')

    def test_net_solar(self):
        self.compare_netcdf_files('net_solar.nc')

    def test_percent_snow(self):
        self.compare_netcdf_files('percent_snow.nc')

    def test_precip(self):
        self.compare_netcdf_files('precip.nc')

    def test_thermal(self):
        self.compare_netcdf_files('thermal.nc')

    def test_wind_speed(self):
        self.compare_netcdf_files('wind_speed.nc')

    def test_wind_direction(self):
        self.compare_netcdf_files('wind_direction.nc')

    def test_snow_density(self):
        self.compare_netcdf_files('snow_density.nc')

    def test_vapor_pressure(self):
        self.compare_netcdf_files('vapor_pressure.nc')
